import argparse

def generate_parser(program_name:str=None):
    program_name = program_name or ""
    parser = argparse.ArgumentParser(
        prog=program_name,
        description=f'PyPendentDrop - {program_name}',
        epilog=f'', add_help=True)
    parser.add_argument('-n', metavar='FILENAME', help='filename', type=argparse.FileType('rb'))
    parser.add_argument('-p', metavar='PXL_DENSITY', help='Pixel density (mm/px)', type=float)
    parser.add_argument('-g', metavar='GRAVITY', help='Acceleration of gravity (typically 9.81)', type=float)
    parser.add_argument('-d', metavar='DELTARHO', help='Density contrast, in kg/L (typically 1.00 for water/air)', type=float)
    parser.add_argument('-o', metavar='OUTPUTFILE', help='Generate graphs [optional:file prefix]', type=str, nargs='?', const='drop', default = None)
    parser.add_argument('-v', help='Verbosity (-v: info, -vv: logger.debug, -vvv: trace)', action="count", default=0)

    group1 = parser.add_argument_group('Drop contour detection options')
    group1.add_argument('-t', metavar='THRESHOLD', help='Threshold level', type=int)
    group1.add_argument('--tlx', help='x position of the top-left corner of the ROI', type=int)
    group1.add_argument('--tly', help='y position of the top-left corner of the ROI', type=int)
    group1.add_argument('--brx', help='x position of the bottom-right corner of the ROI', type=int)
    group1.add_argument('--bry', help='y position of the bottom-right corner of the ROI', type=int)

    group2 = parser.add_argument_group(title='Initial estimation of the parameters',
                                       description='Values of the parameters passed as initial estimation to the optimizer')
    group2.add_argument('--ai', metavar='ANGLE_INIT', help='Angle of gravity (in deg)', type=float)
    group2.add_argument('--xi', metavar='TIP_X_INIT', help='Tip x position (in px)', type=float)
    group2.add_argument('--yi', metavar='TIP_Y_INIT', help='Tip y position (in px)', type=float)
    group2.add_argument('--ri', metavar='R0_INIT', help='Drop radius r0 (in mm)', type=float)
    group2.add_argument('--li', metavar='LCAP_INIT', help='Capillary length lc (in mm)', type=float)

    group3 = parser.add_argument_group('Imposed parameters',
                                       description='Non-free parameters imposed to the optimizer (these are not varied to optimize the fit)')
    group3.add_argument('--af', help='Fix the angle of gravity', action='store_false')
    group3.add_argument('--xf', help='Fix the tip x position', action='store_false')
    group3.add_argument('--yf', help='Fix the tip y position', action='store_false')
    group3.add_argument('--rf', help='Fix the drop radius', action='store_false')
    group3.add_argument('--lf', help='Fix the capillary length', action='store_false')

    return parser


# Main window
from typing import Tuple, Union, Optional, Dict, Any, List
import numpy as np

# from pyqtgraph.Qt.QtGui import QPixmap
from pyqtgraph.Qt.QtWidgets import QMainWindow, QFileDialog

from .. import *

from .mainwindow_ui import Ui_PPD_MainWindow
from .plotwidget import ppd_plotWidget


RAD_PER_DEG = np.pi/180
DEG_PER_RAD = 180/np.pi

class ppd_mainwindow(QMainWindow, Ui_PPD_MainWindow):
    def __init__(self):
        super().__init__()
        
        self.setupUi(self) # we benefit of the double

        ### LEFT SIDE

        # # The widget #0 of displayStackedWidget is the welcome image illustration
        # pixmap = QPixmap('assets/images/ppd_illustration.png')
        # self.welcomeIcon.setPixmap(pixmap)

        # # The widget #1 of displayStackedWidget is the image shown by pyqtgraph
        self.plotWidget = ppd_plotWidget(self)
        self.displayStackedWidget.addWidget(self.plotWidget)

        self.plotWidget.roi.sigRegionChanged.connect(self.ROIMoved)
        self.ROI_TL_x_spinBox.editingFinished.connect(self.ROIChanged)
        self.ROI_TL_y_spinBox.editingFinished.connect(self.ROIChanged)
        self.ROI_BR_x_spinBox.editingFinished.connect(self.ROIChanged)
        self.ROI_BR_y_spinBox.editingFinished.connect(self.ROIChanged)

        self.plotWidget.isoCtrlLine.sigDragged.connect(self.thresholdMoved)
        self.customThresholdSpinBox.editingFinished.connect(self.thresholdChanged)

        self.autoThresholdCheckBox.toggled.connect(self.plotWidget.set_manualIsoCurve_immobile)

        ### RIGHT SIDE

        ### IMAGE AND THRESHOLD TAB
        ## IMAGE GROUPBOX
        self.imageFileBrowsePushButton.clicked.connect(self.choose_image_file)
        self.imageFileLineEdit.editingFinished.connect(self.try_to_load_image)

        ## THRESHOLD GROUPBOX
        # we hide for now the threshold options, might like them later
        self.subpixelCheckBox.setVisible(False)
        self.smoothingCheckBox.setVisible(False)
        self.SmoothingDistanceLabel.setVisible(False)
        self.smoothingDistanceSpinBox.setVisible(False)
        self.autoThresholdCheckBox.toggled.connect(self.autoThresholdToggled)

        ### MEASUREMENT TAB
        ## GUESS + FIT
        self.analysisTabs.setTabEnabled(1, False)
        self.pixelDensitySpinBox.editingFinished.connect(self.pixelDensityChanged)
        self.pixelSizeSpinbox.editingFinished.connect(self.pixelSpacingChanged)
        self.autoGuessPushButton.clicked.connect(self.guessParameters)


        self.anglegSpinBox.valueChanged.connect(self.angleg_manualchange)
        self.tipxSpinBox.valueChanged.connect(self.tipx_manualchange)
        self.tipySpinBox.valueChanged.connect(self.tipy_manualchange)
        self.r0SpinBox.valueChanged.connect(self.r0_manualchange)
        self.caplengthSpinBox.valueChanged.connect(self.caplength_manualchange)

        self.parameters:Parameters = Parameters()
        self.applyParameters()

        self.gSpinBox.valueChanged.connect(self.g_manualchange)
        self.dSpinBox.valueChanged.connect(self.d_manualchange)
        self.g_manualchange()
        self.d_manualchange()

        self.optimizePushButton.clicked.connect(self.optimizeParameters)

    def choose_image_file(self):
        
        dialog = QFileDialog(self)
        
        dialog.setFileMode(QFileDialog.ExistingFile) # single file
        dialog.setViewMode(QFileDialog.Detail)
        ## ADD other images types...
        dialog.setNameFilter("Images (*.png *.tif *.tiff *.jpg *.jpeg)")
        
        if dialog.exec():
            fileName = dialog.selectedFiles()[0]
            debug(f'choose_image_file: Dialog-selected file name: {fileName}')
            self.imageFileLineEdit.setText(fileName)

            self.try_to_load_image()

    def try_to_load_image(self):
        fileName:str = self.imageFileLineEdit.text()
        if self.plotWidget.load_image(filepath=fileName):
            self.displayStackedWidget.setCurrentIndex(1)
            self.analysisTabs.setTabEnabled(1, True)
        else:
            self.displayStackedWidget.setCurrentIndex(0)
            self.analysisTabs.setTabEnabled(1, False)

    ### WORK ON IMAGE

    def ROIMoved(self):
        ROIpos = self.plotWidget.roi.pos()
        ROIsize = self.plotWidget.roi.size()
        RL = [int(ROIpos[0]), int(ROIpos[1])]
        BR = [int(ROIpos[0] + ROIsize[0]), int(ROIpos[1] + ROIsize[1])]
        if self.ROI_TL_x_spinBox.value() != RL[0]:
            self.ROI_TL_x_spinBox.setValue(RL[0])
        if self.ROI_TL_y_spinBox.value() != RL[1]:
            self.ROI_TL_y_spinBox.setValue(RL[1])
        if self.ROI_BR_x_spinBox.value() != BR[0]:
            self.ROI_BR_x_spinBox.setValue(BR[0])
        if self.ROI_BR_y_spinBox.value() != BR[1]:
            self.ROI_BR_y_spinBox.setValue(BR[1])
        self.plotWidget.iso.setROI(ROIpos, ROIsize)
        self.autoThresholdToggled()
    
    def ROIChanged(self):
        ROI_TL_x = self.ROI_TL_x_spinBox.value()
        ROI_TL_y = self.ROI_TL_y_spinBox.value()
        ROI_BR_x = self.ROI_BR_x_spinBox.value()
        ROI_BR_y = self.ROI_BR_y_spinBox.value()
        ROIpos = (ROI_TL_x, ROI_TL_y)
        ROIsize = (ROI_BR_x - ROI_TL_x, ROI_BR_y - ROI_TL_y)
        self.plotWidget.roi.setPos(ROIpos)
        self.plotWidget.roi.setSize(ROIsize)

        self.plotWidget.iso.setROI(ROIpos, ROIsize)
        self.autoThresholdToggled()
        # self.imgShowWidget.iso.offset = np.array([ROIpos[0], ROIpos[1]])
        # self.imgShowWidget.iso.setData(self.imgShowWidget.roi.getArrayRegion(self.imgShowWidget.data, self.imgShowWidget.imageItem))

    def autoThresholdToggled(self, autoThreshold:bool=None):
        if autoThreshold is None:
            autoThreshold = self.autoThresholdCheckBox.isChecked()
        if autoThreshold:
            threshold = auto_threshold(self.plotWidget.iso.data, self.plotWidget.iso.roi)
            self.customThresholdSpinBox.setValue(threshold)
        self.thresholdChanged()

    def thresholdMoved(self):
        level = self.plotWidget.isoCtrlLine.value()
        if self.customThresholdSpinBox.value() != level:
            self.customThresholdSpinBox.setValue(level)
            self.thresholdChanged()
        
    def thresholdChanged(self, threshold:int = None):
        if threshold is None:
            threshold=self.customThresholdSpinBox.value()
        if self.plotWidget.isoCtrlLine.value() != threshold:
            self.plotWidget.isoCtrlLine.setValue(threshold)
        
        if self.plotWidget.iso.level != threshold:
            self.plotWidget.iso.setLevel(threshold)

    ### ESTIMATE PARAMETERS

    def pixelSpacingChanged(self, pixelSpacing:Optional[float]=None):
        if pixelSpacing is None:
            pixelSpacing = self.pixelSizeSpinbox.value()
        debug(f'pixelSpacingChanged with spacing={pixelSpacing} px/mm')
        self.pixelDensitySpinBox.editingFinished.disconnect(self.pixelDensityChanged)

        self.parameters.set_px_spacing(pixelSpacing)
        self.pixelDensitySpinBox.setValue(self.parameters.get_px_density() or 0)

        self.pixelDensitySpinBox.editingFinished.connect(self.pixelDensityChanged)

        self.applyParameters() # actualize r0_mm and lc_mm

        self.autoGuessPushButton.setEnabled(self.parameters.can_estimate())

    def pixelDensityChanged(self, pixelDensity:float=None):
        if pixelDensity is None:
            pixelDensity = self.pixelDensitySpinBox.value()
        debug(f'pixelDensityChanged with density={pixelDensity} px/mm')
        self.pixelSizeSpinbox.editingFinished.disconnect(self.pixelSpacingChanged)

        self.parameters.set_px_density(pixelDensity)
        self.pixelSizeSpinbox.setValue(self.parameters.get_px_spacing() or 0)

        self.pixelSizeSpinbox.editingFinished.connect(self.pixelSpacingChanged)

        self.applyParameters() # actualize r0_mm and lc_mm

        self.autoGuessPushButton.setEnabled(self.parameters.can_estimate())

    def angleg_manualchange(self, angleg:Optional[float]=None):
        if angleg is None:
            angleg = self.anglegSpinBox.value()
        self.parameters.set_a_deg(angleg)
        self.actualizeComputedCurve()

    def tipx_manualchange(self, tipx:Optional[float]=None):
        if tipx is None:
            tipx = self.tipxSpinBox.value()
        self.parameters.set_x_px(tipx)
        self.actualizeComputedCurve()

    def tipy_manualchange(self, tipy:Optional[float]=None):
        if tipy is None:
            tipy = self.tipySpinBox.value()
        self.parameters.set_y_px(tipy)
        self.actualizeComputedCurve()

    def r0_manualchange(self, r0:Optional[float]=None):
        if r0 is None:
            r0 = self.r0SpinBox.value()
        self.parameters.set_r_mm(r0)
        self.actualizeComputedCurve()
        self.actualizeSurfaceTension()

    def caplength_manualchange(self, caplength:Optional[float]=None):
        if caplength is None:
            caplength = self.caplengthSpinBox.value()
        self.parameters.set_l_mm(caplength)
        self.actualizeComputedCurve()
        self.actualizeSurfaceTension()

    def applyParameters(self):
        self.anglegSpinBox.valueChanged.disconnect(self.angleg_manualchange)
        self.tipxSpinBox.valueChanged.disconnect(self.tipx_manualchange)
        self.tipySpinBox.valueChanged.disconnect(self.tipy_manualchange)
        self.r0SpinBox.valueChanged.disconnect(self.r0_manualchange)
        self.caplengthSpinBox.valueChanged.disconnect(self.caplength_manualchange)

        self.anglegSpinBox.setValue(self.parameters.get_a_deg() or 0)
        self.tipxSpinBox.setValue(self.parameters.get_x_px() or 0)
        self.tipySpinBox.setValue(self.parameters.get_y_px() or 0)
        self.r0SpinBox.setValue(self.parameters.get_r_mm() or 0)
        self.caplengthSpinBox.setValue(self.parameters.get_l_mm() or 0)

        self.anglegSpinBox.valueChanged.connect(self.angleg_manualchange)
        self.tipxSpinBox.valueChanged.connect(self.tipx_manualchange)
        self.tipySpinBox.valueChanged.connect(self.tipy_manualchange)
        self.r0SpinBox.valueChanged.connect(self.r0_manualchange)
        self.caplengthSpinBox.valueChanged.connect(self.caplength_manualchange)

        self.actualizeComputedCurve()
        self.actualizeSurfaceTension()

    def guessParameters(self):
        if self.parameters.can_estimate():
            px_per_mm = self.parameters.get_px_density()
            threshold = self.customThresholdSpinBox.value()

            mainContour = self.plotWidget.isoCurve_level(level=threshold)

            self.parameters = estimate_parameters(np.array(self.plotWidget.iso.data), mainContour, px_per_mm=px_per_mm)
            self.g_manualchange() ; self.d_manualchange() # update with the values of physical parameters
            self.parameters.describe(printfn=info, descriptor='estimated')

            self.applyParameters()

    ### OPTIMIZE PARAMETERS

    def canComputeProfile(self) -> bool:
        canDoOptimization = self.parameters.can_optimize()

        self.optimizePushButton.setEnabled(canDoOptimization)
        if not(canDoOptimization):
            self.gammaSpinBox.setValue(0)
            self.bondSpinBox.setValue(0)
        return canDoOptimization

    def optimizeParameters(self):
        if self.canComputeProfile():

            threshold = self.customThresholdSpinBox.value()
            mainContour = self.plotWidget.isoCurve_level(level=threshold)

            to_fit=[self.anglegCheckBox.isChecked(),
                    self.tipyCheckBox.isChecked(),
                    self.tipxCheckBox.isChecked(),
                    self.r0CheckBox.isChecked(),
                    self.caplengthCheckBox.isChecked()]
            trace(f'optimizeParameters: to_fit={to_fit} (from checkboxes)')

            opti_success, self.parameters = optimize_profile(mainContour, parameters_initialguess=self.parameters, to_fit=to_fit)
            self.parameters.describe(printfn=info, descriptor='optimized')

            self.applyParameters()

    def actualizeComputedCurve(self):
        if self.parameters.can_show_tip_position():
            self.plotWidget.scatter_droptip(self.parameters.get_xy_px())
        else:
            self.plotWidget.hide_scatter_droptip()
        if self.canComputeProfile():
            R, Z = integrated_contour(self.parameters)

            self.plotWidget.plot_computed_profile(R, Z)
        else:
            self.plotWidget.hide_computed_profile()

    ### PHYSICS
    def g_manualchange(self, g:Optional[float]=None):
        if g is None:
            g = self.gSpinBox.value()
        self.parameters.set_g(g)
        self.actualizeSurfaceTension()
    def d_manualchange(self, d:Optional[float]=None):
        if d is None:
            d = self.dSpinBox.value()
        self.parameters.set_d(d)
        self.actualizeSurfaceTension()

    def actualizeSurfaceTension(self):
        if self.canComputeProfile():
            self.bondSpinBox.setValue(self.parameters.get_bond() or 0)
            self.gammaSpinBox.setValue(self.parameters.get_surface_tension_mN() or 0)



        






import numpy as np
from typing import Tuple, Union, Optional, Dict, Any, List
import pyqtgraph as pg
from pyqtgraph.Qt.QtCore import QRectF
from pyqtgraph.Qt import QtGui

from .. import *

# Interpret image data as row-major instead of col-major
pg.setConfigOptions(imageAxisOrder='row-major')
pg.setConfigOption('background', (255,255,255, 100))


# Custom Isocurve item with an offset
class IsocurveItemWithROI(pg.IsocurveItem):

    def __init__(self, offset=None):
        super().__init__(self)
        self.roi = [0,0, None, None] # TLx, TLy, BRx, BRy
        self.imgheight, self.imgwidth = None, None

    def setROI(self, ROIpos, ROIsize):
        self.roi = [int(ROIpos[0]), int(ROIpos[1]),
                    int(ROIpos[0]+ROIsize[0]), int(ROIpos[1]+ROIsize[1])] # TLx, TLy, BRx, BRy
        trace(f'isoCurve: setROI: roi is now {self.roi}')

        self.generatePath()
        self.update()

    def generatePath(self):
        if self.data is None:
            self.path = None
            return
        
        # if self.axisOrder == 'row-major':
        #     data = self.data.T
        # else:
        #     data = self.data
        data = self.data

        # lines = pg.isocurve(data, self.level, connected=True, extendToEdge=True)
        lines = detect_contourlines(data, self.level, roi=self.roi)
        self.path = QtGui.QPainterPath()
        for line in lines:
            self.path.moveTo(*line[0])
            for p in line[1:]:
                self.path.lineTo(*p)

class ppd_plotWidget(pg.GraphicsLayoutWidget):
    def __init__(self, parentWidget):
        super().__init__(parentWidget)

        # A plot area (ViewBox + axes) for displaying the image
        self.plot = self.addPlot(title="")
        self.plot.autoRange()

        self.defaultLevel:float = 127
        
        # Item for displaying image data
        self.imageItem = pg.ImageItem()
        self.plot.addItem(self.imageItem)
        self.imageItem.getViewBox().invertY(True)
        self.imageItem.getViewBox().setAspectLocked(lock=True, ratio=1)
        self.imageItem.getViewBox().enableAutoRange()
        
        self.load_image(filepath=None)
        self.computedProfilePlotDataItem = None
        self.dropTipPlotDataItem = None
        
        self._setupIsocurve()
        self._setupROI()
        self._setupHistogram()
        
        # self.resize(800, 800)
        # self.show()
        
    def load_image(self, filepath:str = None) -> bool:
        """Tries to load an image to the widget

        Parameters
        ----------
        filepath

        Returns
        -------
        success

        """
        success, self.data = import_image(filepath)
            
        self.imgheight, self.imgwidth = self.data.shape
        self.imageItem.clear()
        self.imageItem.setImage(self.data)
        # self.imageItem.getViewBox().setLimits(xMin=0-max(0, (height-width)/2), xMax=width+max(0, (height-width)/2), 
        #                                       yMin=0-max(0, (width-height)/2), yMax=height+max(0, (width-height)/2))
        
        if success:
            self.iso.setData(self.data, level=self.defaultLevel)

            ROIpos = [self.imgwidth*(1-self.initialROIfract)/2, self.imgheight*(1-self.initialROIfract)/2]
            ROIsize = [self.imgwidth*self.initialROIfract, self.imgheight*self.initialROIfract]
            ROIpos = np.array(ROIpos).astype(int)
            ROIsize = np.array(ROIsize).astype(int)
            trace(f'plotWidget: load_image: ROI position = {ROIpos} | ROI size = {ROIsize}')
            ROImaxBounds = QRectF(0-.5, 0-.5, self.imgwidth+1, self.imgheight+1)
            self.roi.maxBounds = ROImaxBounds
            self.roi.setPos(ROIpos)
            self.roi.setSize(ROIsize)

            self.iso.setData(self.data, level=self.defaultLevel)
            self.iso.setROI(ROIpos, ROIsize)
            self.isoCtrlLine.setValue(self.defaultLevel)

            self.hist.setLevels(self.data.min(), self.data.max())

            if self.computedProfilePlotDataItem is not None:
                self.computedProfilePlotDataItem.setVisible(False)

        return success
        
    def _setupIsocurve(self, level=None):
        if level is None:
            level = self.defaultLevel
        ### ISOCURVE
        self.iso:IsocurveItemWithROI = IsocurveItemWithROI()
        self.iso.setPen(pg.mkPen(color='g', width=2))
        self.iso.setLevel(level)
        self.iso.setParentItem(self.imageItem)
        self.iso.setZValue(5)
        # We can also use smoothed data for the isoline, giving smoother results:
        # self.iso.setData(pg.gaussianFilter(data, (2, 2)))
        # Here we decide to use the raw data
        self.iso.setData(self.data) 
    
    def _setupROI(self):        
        # Custom ROI for selecting an image region
        self.initialROIfract = .95
        ROIpos = [self.imgwidth*(1-self.initialROIfract)/2, self.imgheight*(1-self.initialROIfract)/2]
        ROIsize = [self.imgwidth*self.initialROIfract, self.imgheight*self.initialROIfract]
        ROIpen = pg.mkPen(color='y', width=1)
        ROIhoverPen = pg.mkPen(color='b', width=1)
        ROIhandlePen = pg.mkPen(color='y', width=2)
        ROIhandleHoverPen = pg.mkPen(color=QtGui.QColor('cyan'), width=2)
        ROImaxBounds = QRectF(0-.5, 0-.5, self.imgwidth+1, self.imgheight+1)
        ROImaxBounds = QRectF(0-.5, 0-.5, self.imgwidth+1, self.imgheight+1)
                
        self.roi = pg.ROI(ROIpos, 
                          ROIsize, 
                          rotatable=False, resizable = False, 
                          pen=ROIpen, hoverPen=ROIhoverPen, 
                          handlePen = ROIhandlePen, handleHoverPen = ROIhandleHoverPen,
                          maxBounds=ROImaxBounds)
        # Side handles
        self.roi.addScaleHandle([0.5, 1], [0.5, 0.])
        self.roi.addScaleHandle([0.5, 0], [0.5, 1.])
        self.roi.addScaleHandle([1, 0.5], [0, 0.5])
        self.roi.addScaleHandle([0, 0.5], [1, 0.5])
        
        # corner handles
        self.roi.addScaleHandle([0, 0], [1, 1])
        self.roi.addScaleHandle([0, 1], [1, 0])
        self.roi.addScaleHandle([1, 1], [0, 0])
        self.roi.addScaleHandle([1, 0], [0, 1])
        
        self.roi.setZValue(10)  # make sure ROI is drawn above image
        self.plot.addItem(self.roi)
        
        # self.roi.sigRegionChanged.connect(self.updateROI)
        # self.updateROI()

    def _setupHistogram(self):
        # Contrast/color control with histogram
        self.hist = pg.HistogramLUTItem()
        self.hist.setImageItem(self.imageItem)
        self.hist.vb.setMouseEnabled(y=False) # disable axis movement
        self.hist.setHistogramRange(0, 255)
        self.hist.region.setBounds((0, 255))
        self.hist.vb.setLimits(yMin=0-5, yMax=255+5)
        self.hist.setLevels(self.data.min(), self.data.max())
        
        self.addItem(self.hist)
        
        # Draggable line for setting isocurve level
        self.isoCtrlLine = pg.InfiniteLine(angle=0, movable=True, pen=pg.mkPen(color='g', width=2))
        self.isoCtrlLine.setValue(self.iso.level)
        self.isoCtrlLine.setZValue(1000) # bring iso line above contrast controls
        self.isoCtrlLine.setMovable(False) # disabled by default
        # self.isoCtrlLine.sigDragged.connect(self.updateIsocurve)
        
        self.hist.vb.addItem(self.isoCtrlLine)

    def set_manualIsoCurve_immobile(self, immobile:bool):
        self.isoCtrlLine.setMovable(not(immobile))

    def disable_isoCurve(self):
        self.iso.visible(False)
        
    def enable_isoCurve(self):
        self.iso.visible(False)
        
    def isoCurve_level(self, level:Optional[float]=None):
        if level is None:
            level = self.defaultLevel
        self.isoCtrlLine.setValue(level)
        return detect_main_contour(self.iso.data, level, roi=self.iso.roi)

    def plot_computed_profile(self, x, y):

        # imv_v = self.imageItem.getView()
        if self.computedProfilePlotDataItem is None:
            # todo: replace this PlotCurveItem by a PlotDataItem (pyqtgraph recomends doing so)
            self.computedProfilePlotDataItem = pg.PlotCurveItem(x=x, y=y, pen=pg.mkPen(color='r', width=2))
            self.plot.addItem(self.computedProfilePlotDataItem)
        else:
            self.computedProfilePlotDataItem.setData(x=x, y=y)
            self.computedProfilePlotDataItem.setVisible(True)

    def hide_computed_profile(self):
        if self.computedProfilePlotDataItem is not None:
            self.computedProfilePlotDataItem.setVisible(False)

    def scatter_droptip(self, xy:Tuple[float, float]):
        if self.dropTipPlotDataItem is None:
            self.dropTipPlotDataItem = pg.PlotDataItem(x=[xy[0]], y=[xy[1]], symbol='o',
                                                       symbolPen=pg.mkPen((0, 255, 0), width=3), symbolBrush=(0, 0, 0))

            self.dropTipPlotDataItem.setZValue(100)
            self.plot.addItem(self.dropTipPlotDataItem)
        else:
            self.dropTipPlotDataItem.setData(x=[xy[0]], y=[xy[1]])
            self.dropTipPlotDataItem.setVisible(True)
    def hide_scatter_droptip(self):
        if self.dropTipPlotDataItem is not None:
            self.dropTipPlotDataItem.setVisible(False)





































__version__ = "0.1.4"

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler()) # so that messages are not shown in strerr
def error(msg:str):
    logger.error(msg)
def warning(msg:str):
    logger.warning(msg)
def info(msg:str):
    logger.info(msg)
def debug(msg:str):
    logger.debug(msg)
def trace(msg:str):
    try:
        logger.trace(msg)
    except:
        pass

debug(f'pypendentdrop version {__version__} loaded')

###### ANALYZE
from .analysis.fetchimage import *
from .analysis.getcontour import *
from .analysis.findparameters import *




### This is pasted forn PyQtGraphs
import os, sys

PYSIDE = 'PySide'
PYSIDE2 = 'PySide2'
PYSIDE6 = 'PySide6'
PYQT4 = 'PyQt4'
PYQT5 = 'PyQt5'
PYQT6 = 'PyQt6'

QT_LIB = os.getenv('PYQTGRAPH_QT_LIB')

if QT_LIB is not None:
    try:
        __import__(QT_LIB)
    except ModuleNotFoundError:
        raise ModuleNotFoundError(f"Environment variable PYQTGRAPH_QT_LIB is set to '{os.getenv('PYQTGRAPH_QT_LIB')}', but no module with this name was found.")

## Automatically determine which Qt package to use (unless specified by
## environment variable).
## This is done by first checking to see whether one of the libraries
## is already imported. If not, then attempt to import in the order
## specified in libOrder.
if QT_LIB is None:
    libOrder = [PYQT6, PYSIDE6, PYQT5, PYSIDE2]

    for lib in libOrder:
        if lib in sys.modules:
            QT_LIB = lib
            break

if QT_LIB is None:
    for lib in libOrder:
        qt = lib + '.QtCore'
        try:
            __import__(qt)
            QT_LIB = lib
            break
        except ImportError:
            pass

if QT_LIB is None:
    raise ImportError("PyPendentDrop GUI requires one of PyQt5, PyQt6, PySide2 or PySide6; none of these packages could be imported (try `pip install PyQt5`).")


try:
    pg = 'pyqtgraph'
    __import__(pg)
except ImportError:
    raise ImportError("PyPendentDrop GUI requires pyqtgraph; which could not be imported (try `pip install pyqtgraph`).")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from . import *
import sys
from . import logfacility
from .ppd_parse import generate_parser

def is_valid_float(element: any) -> bool:
    if element is None:
        return False
    try:
        if float(element) < 0:
            warning('The number cannot be negative !?')
            return False
        return True
    except ValueError:
        warning(f'This is not a number: {element}')
        return False
def main():
    parser = generate_parser('Command Line Interface')

    args = parser.parse_args()

    logfacility.set_verbose(args.v)

    ### Getting the image
    imagefile = args.n
    logger.debug(f'Image path provided: {imagefile}')
    if imagefile is None:
        logger.info(f'No image file provided.')
        logger.info(f'You can use the -n option to specify the image you want to analyze.')
        logger.info(f'Example: ppd-cli -n image.tif)')

    import_success, img = import_image(imagefile)
    while not import_success:
        if imagefile is None:
            print('No image file provided.')
        print('Please provide a valid path for the image you want to analyze')
        imagefile:str = str(input('Image file path: '))
        logger.debug(f'Image path provided: {imagefile}')
        import_success, img = import_image(imagefile)

    ### Getting the pixel density
    px_per_mm = args.p
    logger.debug(f'Pixel density provided: {px_per_mm} px/mm')
    if px_per_mm is None:
        logger.info(f'No pixel density provided.')
        logger.info(f'You can use the -p option specify the pixel density, in px/mm.')
        logger.info(f'Example: ppd-cli -p 57.0)')

    while not(is_valid_float(px_per_mm)):
        if px_per_mm is None:
            print('No pixel density provided.')
        print('Please provide a valid pixel density for the image you want to analyze')
        px_per_mm:str = str(input('Pixel density (px/mm): '))
        logger.debug(f'Pixel density provided: {px_per_mm} px/mm')
    px_per_mm:float = float(px_per_mm)

    import_success, img = import_image(imagefile)
    if import_success:
        logger.debug(f'Import image successful.')
    else:
        logger.error(f'Could not retreive the image at {imagefile}')
        sys.exit(200)

    height, width = img.shape
    logger.debug(f'Image shape: {height}x{width}')

    roi = format_roi(img, [args.tlx, args.tly, args.brx, args.bry])
    logger.debug(f'roi = {roi}')

    threshold = args.t
    if threshold is None:
        logger.debug('Threshold not provided, using auto_threshold to provide it.')
        threshold = auto_threshold(img, roi=roi)

    logger.debug(f'Threshold level: {threshold}')

    lines = detect_contourlines(img, threshold, roi=roi)
    linelengths = [len(line) for line in lines]

    logger.debug(f'Number of lines: {len(lines)}, lengths: {linelengths}')

    cnt = detect_main_contour(img, threshold, roi=roi)

    logger.debug(f'Drop contour: {cnt.shape[1]} points')

    estimated_parameters = estimate_parameters(img, cnt, px_per_mm)

    args_parameters = Parameters()
    args_parameters.set_px_density(px_per_mm)
    args_parameters.set_a_deg(args.ai)
    args_parameters.set_x_px(args.xi)
    args_parameters.set_y_px(args.yi)
    args_parameters.set_r_mm(args.ri)
    args_parameters.set_l_mm(args.li)
    args_parameters.describe(printfn=trace, descriptor='from arguments')

    initial_parameters = Parameters()
    initial_parameters.set_px_density(px_per_mm)
    initial_parameters.set_a_deg(args.ai or estimated_parameters.get_a_deg())
    initial_parameters.set_x_px(args.xi or estimated_parameters.get_x_px())
    initial_parameters.set_y_px(args.yi or estimated_parameters.get_y_px())
    initial_parameters.set_r_mm(args.ri or estimated_parameters.get_r_mm())
    initial_parameters.set_l_mm(args.li or estimated_parameters.get_l_mm())
    initial_parameters.describe(printfn=debug, descriptor='initial')

    logger.debug(f'chi2: {compute_gap_dimensionless(cnt, parameters=initial_parameters)}')


    if args.o is not None:
        from . import ppdplot

        ppdplot.generate_figure(img, cnt, parameters=initial_parameters,
                             prefix=args.o, comment='estimated parameters', suffix='_initialestimate', filetype='pdf', roi=roi)
    else:
        pass
        # info('You did not ask for graphs to be generated')

    to_fit = [args.af, args.xf, args.yf, args.rf, args.lf]

    logger.debug(f'to_fit: {to_fit}')

    opti_success, opti_params = optimize_profile(cnt, parameters_initialguess=initial_parameters, to_fit=to_fit,
                                                 method=None)

    if not opti_success:
        logger.warning('Optimization failed :( Falling back to the estimated parameters.')
        sys.exit(-2)
    else:
        opti_params.describe(printfn=info, descriptor='optimized')

        logger.debug(f'chi2: {compute_gap_dimensionless(cnt, parameters=opti_params)}')

        print(f'Capillary length: {round(opti_params.get_l_mm(), 3)} mm')
        print(f'Bond number: {round(opti_params.get_bond(), 3)}')

        ### Getting the contrast density
        g = args.g
        logger.debug(f'Gravity acceleration provided: {g} m/s2')
        if g is None:
            logger.info(f'No gravity acceleration provided.')
            logger.info(f'You can use the -g option specify the gravity acceleration (in m/s^2).')
            logger.info(f'Example (for Paris): ppd-cli -g 9.81)')
        while not(is_valid_float(g)):
            if g is None:
                print(f'No gravity acceleration provided.')
            print('Please provide a valid gravity acceleration for the image you want to analyze')
            g:str = str(input('Gravity acceleration (m/s^2): '))
            logger.debug(f'Gravity acceleration provided: {g} m/s2')
        g:float = float(g)
        opti_params.set_g(g)


        d = args.d
        logger.debug(f'Density contrast provided: {d} kg/L')
        if d is None:
            logger.info(f'No density contrast provided.')
            logger.info(f'You can use the -d option specify the density contrast (density difference between the fluids, in kg/L).')
            logger.info(f'Example (for water in air): ppd-cli -d 1.00)')
        while not(is_valid_float(d)):
            if d is None:
                print('No density contrast provided.')
            print('Please provide a valid density contrast for the image you want to analyze')
            d:str = str(input('Density contrast (kg/L): '))
            logger.debug(f'Density contrast provided: {d}')
        d:float = float(d)
        opti_params.set_d(d)

        print(f'Surface tension gamma: {round(opti_params.get_surface_tension_mN(), 3)} mN/m')

        if args.o is not None:
            try:
                from . import ppdplot
                ppdplot.generate_figure(img, cnt, parameters=opti_params,
                                     prefix=args.o, comment='optimized parameters', suffix='_optimalestimate', filetype='pdf', roi=roi)
            except ImportError:
                raise ImportError("Using the `-o` option requires matplotlib; which could not be imported (try `pip install matplotlib`).")

    sys.exit(0)


import logging
from . import logger

def trace(msg:str):
    if hasattr(logging, 'TRACE'):
        logging.getLogger(__name__).trace(msg)

def addLoggingLevel(levelName, levelNum, methodName=None):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present
    """
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName):
        raise AttributeError('{} already defined in logging module'.format(levelName))
    if hasattr(logging, methodName):
        raise AttributeError('{} already defined in logging module'.format(methodName))
    if hasattr(logging.getLoggerClass(), methodName):
        raise AttributeError('{} already defined in logger class'.format(methodName))

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def logForLevel(self, message, *args, **kwargs):
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)
    def logToRoot(message, *args, **kwargs):
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)
addLoggingLevel('TRACE', logging.DEBUG - 5)

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

class CustomFormatter(logging.Formatter):
    format = "[%(levelname)s] - %(message)s"

    FORMATS = {
        logging.TRACE: bcolors.OKBLUE + format + bcolors.ENDC,
        logging.DEBUG: bcolors.OKCYAN + format + bcolors.ENDC,
        logging.INFO: bcolors.OKGREEN + format + bcolors.ENDC,
        logging.WARNING: bcolors.WARNING + format + bcolors.ENDC,
        logging.ERROR: bcolors.FAIL + format + bcolors.ENDC,
        logging.CRITICAL: bcolors.FAIL + format + bcolors.ENDC
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)

def set_verbose(verbose:int):
    level = logging.WARNING
    if verbose == 1:
        level=logging.INFO
    elif verbose == 2:
        level=logging.DEBUG
    elif verbose > 2:
        level=logging.TRACE


    logger.setLevel(level)

    ch = logging.StreamHandler()
    ch.setLevel(level)

    ch.setFormatter(CustomFormatter())

    logger.addHandler(ch)


from . import *

try:
    mpl = 'matplotlib'
    __import__(mpl)
except ImportError:
    raise ImportError("PyPendentDrop plotting module requires matplotlib; which could not be imported (try `pip install matplotlib`).")

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import patches
from scipy.integrate import trapezoid

plt.rcParams["figure.figsize"] = (8, 6)
plt.rcParams["figure.max_open_warning"] = 50

# plt.rcParams['pgf.texsystem'] = 'pdflatex' # use this if you have LaTeX
plt.rcParams.update({'font.family': 'serif', 'font.size': 10,
                     'figure.titlesize' : 10,
                     'axes.labelsize': 10,'axes.titlesize': 12,
                     'legend.fontsize': 10})

from .analysis.getcontour import format_roi

def plot_roi(ax, image:np.ndarray, roi=None, **kwargs):
    tlx, tly, brx, bry = format_roi(image, roi)
    if kwargs.get('color', None) is None:
        kwargs['color'] = 'y'
    if kwargs.get('lw', None) is None and kwargs.get('linewidth', None) is None:
        kwargs['lw'] = 2
    if kwargs.get('ls', None) is None and kwargs.get('linestyle', None) is None:
        kwargs['ls'] = '--'
    if kwargs.get('label', None) is None:
        kwargs['label'] = 'ROI'
    ax.plot([tlx, tlx, brx, brx, brx], [tly, bry, bry, tly, tly], **kwargs)

def plot_contour(ax, contour, **kwargs):
    xcontour, ycontour = contour[0], contour[1]
    if kwargs.get('lw', None) is None and kwargs.get('linewidth', None) is None:
        kwargs['lw'] = 1.5
    if kwargs.get('ls', None) is None and kwargs.get('linestyle', None) is None:
        kwargs['ls'] = '-'
    if kwargs.get('label', None) is None:
        kwargs['label'] = 'Contour'
    ax.plot(xcontour, ycontour, **kwargs)

def plot_image_contour(ax, image:np.ndarray, contour:np.ndarray, parameters:Parameters, comment='', roi=None):
    roi = format_roi(image, roi)
    ax.set_title(f'Drop image and contour ({comment})')
    ax.imshow(image, cmap='gray', aspect='equal')
    plot_roi(ax, image, roi)

    plot_contour(ax, contour, color='lime', label='Detected contour')

    gravity_angle = parameters.get_a_rad()
    x_tip_position, y_tip_position = parameters.get_xy_px()

    l = max(image.shape)
    ax.plot((x_tip_position + l * np.sin(-gravity_angle), x_tip_position - l * np.sin(-gravity_angle)), (y_tip_position - l * np.cos(-gravity_angle), y_tip_position + l * np.cos(-gravity_angle)),
            color='b', lw=2, ls='--', label=f'Direction of gravity')

    ax.scatter(x_tip_position, y_tip_position, s=50, fc='k', ec='lime', linewidths=2, label=f'Tip position', zorder=4)

    r0_px = parameters.get_r_px()

    drop_center_x = x_tip_position + r0_px * np.sin(-gravity_angle)
    drop_center_y = y_tip_position - r0_px * np.cos(-gravity_angle)
    e1 = patches.Arc((drop_center_x, drop_center_y), 2 * r0_px, 2 * r0_px,  # WARNING CONVENTION
                     theta1 = 0 - gravity_angle*180/np.pi, theta2 = 180 - gravity_angle*180/np.pi,
                     linewidth=2, fill=False, zorder=2, color='darkred', ls='--', label=f'Curvature')
    ax.add_patch(e1)

    contour_integrated = integrated_contour(parameters)

    plot_contour(ax, contour_integrated, color='r', label='Computed profile')

    ax.legend()
    ax.set_xlabel('x [px]')
    ax.set_xlim(0, image.shape[1])
    ax.set_ylabel('y [px]')
    ax.set_ylim(image.shape[0], 0)

def plot_difference(axtop, axbot, contour, parameters:Parameters, comment=''):
    # axtop.set_title(f'chi2: {analyze.compute_gap_dimensionless(fitparams, contour, px_per_mm=px_per_mm)}')
    axtop.set_title(f'Comparison of detected contour and computed profile')

    fitparams:Fitparams = parameters.get_fitparams()

    gravity_angle, x_tip_position, y_tip_position, r0_px, capillary_length_px = fitparams

    tipRadius = r0_px / capillary_length_px

    # hence the profile
    R, Z = compute_nondimensional_profile(tipRadius)

    # FOR COMPUTE THE DIFF : we take it backward
    XY = contour.copy()

    #moving
    XY[0] -= x_tip_position
    XY[1] -= y_tip_position

    #  rotating and scaling
    XY = rotate_and_scale(XY, angle=-gravity_angle, scalefactor=-1 / capillary_length_px)

    # # cutting off :
    # XY = np.take(XY, np.where(XY[1] < Z.max())[0], axis=1)


    # separating the two sides
    rightside = XY[0] > 0
    X1, Y1 = np.take(XY, np.where(rightside)[0], axis=1)
    # X2, Y2 = -X[X < 0], Y[X < 0]
    # X2, Y2 = XY[:, np.bitwise_not(rightside)
    X2, Y2 = np.take(XY, np.where(np.bitwise_not(rightside))[0], axis=1)
    X2 *= -1

    # the differences
    R1 = np.interp(Y1, Z, R) # the radius corresponding to the side 1
    R2 = np.interp(Y2, Z, R) # the radius corresponding to the side 2

    R1[Y1 > Z.max()] = R[np.argmax(Z)]
    R2[Y2 > Z.max()] = R[np.argmax(Z)]

    R1[Y1 < Z.min()] *= 0
    R2[Y2 < Z.min()] *= 0
    DX1 = X1 - R1
    DX2 = X2 - R2

    difference = np.abs(trapezoid(DX1**2, Y1)) + np.abs(trapezoid(DX2**2, Y2))
    # print(f'DGB: CHI2: {chi2}')

    ### AX ON TOP

    axtop.plot(Z, R, c='m', ls=':', lw=1)
    axtop.plot(Y1, R1, c='r', lw=1)
    axtop.plot(Y2, R2, c='r', lw=1, label=f'Computed contour ({comment})')
    axtop.plot(Y1, X1, c='lime', lw=1, ls='--', label='Drop contour (right side)')
    axtop.plot(Y2, X2, c='lime', lw=1, ls=':', label='Drop contour (left side)')

    axtop.legend()
    axtop.yaxis.tick_right()
    axtop.yaxis.set_label_position('right')
    axtop.set_xlim(min(Y1.min(), Y2.min(), 0), max(Y1.max(), Y2.max()))
    axtop.set_ylim(0, max(X1.max(), X2.max(), R.max()) * 1.2)
    axtop.set_xlabel('Z [dimensionless]')
    axtop.set_ylabel('R [dimensionless]')

    ### AX ON BOTTOM

    axbot.axhline(0, c='gray', alpha=.3)
    axbot.plot(Y1, DX1, ls='--', c='gray', label='Right side')
    axbot.plot(Y2, DX2, ls=':', c='gray', label='Left side')

    axbot.legend()
    axbot.yaxis.tick_right()
    axbot.yaxis.set_label_position('right')
    axbot.set_xlim(min(Y1.min(), Y2.min(), 0), max(Y1.max(), Y2.max()))
    bnd = max(np.abs(DX1).max(), np.abs(DX2).max()) * 1.2
    axbot.set_ylim(-bnd, bnd)
    axbot.set_xlabel('Z [dimensionless]')
    axbot.set_ylabel('Detected contour - computed profile')
    # axbot.spines['right'].set_color('gray')
    # axbot.tick_params(axis='y', colors='gray')
    # axbot.xaxis.label.set_color('gray')

def generate_figure(data, contour, parameters:Parameters, prefix=None, comment=None, suffix=None, filetype='pdf', roi=None):
    if prefix is None:
        prefix= ''
    if suffix is None:
        suffix=''
    if comment is None:
        comment=''
    plt.figure()
    ax = plt.subplot(1, 2, 1)
    plot_image_contour(ax, data, contour, parameters, comment=comment, roi=roi)
    ax1, ax2 = plt.subplot(2, 2, 2), plt.subplot(2, 2, 4)
    plot_difference(ax1, ax2, contour, parameters)
    prefix = f'{prefix}{suffix}'
    plt.savefig(prefix + '.' + filetype, dpi=300)


from typing import Tuple, Union, Optional, Dict, Any, List
import math
import numpy as np
from scipy.integrate import odeint, trapezoid
from scipy.optimize import minimize
import time

from .. import error, warning, info, debug, trace

Fitparams = List[float]
def roundifnotnone(val:Optional[float], digits=2, unit=None):
    return None if val is None else f"{round(val, digits)}{'' if unit is None else ' '+unit}"
class Parameters():
    """The parameters describing a drop.

    This class stores the parameters describing the drop in the image
    (gravity angle, tip position in px, radius of curvation in px, capillary length in px).
    It also stores physically relevant informations, such as the pixel-to-mm conversion constant, the acceleration of gravity, the density ratio between the two fluids.

    It can return all these parameters and equivalent ondes (bond number, surface tension) in different units.

    Use :meth:`describe <pypendentdrop.Parameters.describe>` to print the values of the parameters.

    *This class has no public attributes, you should use the (numerous) getter and setter methods to manipulate the variables.*

    This class has str and repr methods #todo add exemples of what this means here.
    """
    RAD_PER_DEG = np.pi/180
    DEG_PER_RAD = 180/np.pi

    def __init__(self):
        self._a_rad: Optional[float] = None
        self._x_px: Optional[float] = None
        self._y_px: Optional[float] = None
        self._r_px: Optional[float] = None
        self._l_px: Optional[float] = None

        self._px_per_mm: Optional[float] = None
        self._d: Optional[float] = 1.00 # delta rho in kg/l
        self._g: Optional[float] = 9.81 # g in m/s2

    def __repr__(self) -> str:
        return f""" 
        pixel density: {roundifnotnone(self.get_px_density(), digits=2, unit='px/mm')}
        gravity: {roundifnotnone(self.get_gravityacc(), digits=3)}
        Delta rho: {roundifnotnone(self.get_densitycontrast(), digits=3)}
        -----
        gravity_angle: {roundifnotnone(self.get_a_deg(), digits=2, unit='deg')} ({roundifnotnone(self.get_a_rad(), digits=4, unit='rad')})
        x_tip_position: {roundifnotnone(self.get_x_px(), digits=2, unit='px')}
        y_tip_position: {roundifnotnone(self.get_y_px(), digits=2, unit='px')}
        r0: {roundifnotnone(self.get_r_mm(), digits=4, unit='mm')} ({roundifnotnone(self.get_r_px(), digits=2, unit='px')})
        capillary_length: {roundifnotnone(self.get_l_mm(), digits=4, unit='mm')} ({roundifnotnone(self.get_l_px(), digits=2, unit='px')})
        """
    def __str__(self) -> str:
        return f"(px_per_mm={self.get_px_density()} ; g={self.get_g()} ; Deltarho={self.get_d()} | a={self.get_a_deg()} deg; x={self.get_x_px()} px; y={self.get_y_px()} px; r={self.get_r_mm()} mm; l={self.get_l_mm()} mm)"

    def describe(self, printfn=print, descriptor = None) -> None:
        """Prints the parameters in the console in a human-friendly fashion.

        Parameters
        ----------
        printfn : function
            The function that will be used to print. By default, print, but you can feed it a logging or debug function.
        descriptor : str
            A short description of this parameter object ('estimated', 'guessed', 'optimized' ...)

        """
        """"""
        printfn(f"Parameters {'' if descriptor is None else ('(' + descriptor + ')')}" + repr(self))

    ###### PIXEL SPACING / SIZE / DENSITY
    def get_px_density(self) -> float:
        """Get the density of pixels, in px/mm. Use ``set_px_density`` to set it.
        
        The pixel density corresponds to the quantity of pixels per unit length in your image.
        For example, if in your image 5 pixels span a length corresponding to 2 mm, the pixel density is 5/2=2.5 pixel per millimeter.
        
        The higher the pixel density, the more zoomed the image is, and in general the better the measurement. 
        A typical pixel density of 100 or higher is recommended for an optimal measurement.
        """
        return self._px_per_mm
    def set_px_density(self, pixel_density:float) -> None:
        self._px_per_mm = pixel_density
    def get_px_spacing(self) -> float:
        """Get the spacing between pixels, in mm. Use ``set_px_spacing`` to set it.
        
        The pixel spacing is to the physical distance corresponding to the space between two adjacent pixels in your image.
        For example, if in your image 5 pixels span a length corresponding to 2 mm, the pixel spacing is 2/5=0.4 millimeter per pixel.
        """
        return None if ((self._px_per_mm or 0) == 0) else 1/self._px_per_mm
    def set_px_spacing(self, pixel_spacing:float) -> None:
        self._px_per_mm = None if ((pixel_spacing or 0) == 0) else 1/pixel_spacing

    ### ANGLE OF GRAVITY SHORT VERSION
    def get_a_rad(self) -> float:
        return self._a_rad
    def set_a_rad(self, gravity_angle_rad:float) -> None:
        self._a_rad = gravity_angle_rad
    def get_a_deg(self) -> float:
        return None if self._a_rad is None else self._a_rad * self.DEG_PER_RAD
    def set_a_deg(self, gravity_angle_deg:float) -> None:
        self._a_rad = None if gravity_angle_deg is None else gravity_angle_deg * self.RAD_PER_DEG

    ### ANGLE OF GRAVITY LONG VERSION
    def get_angleofgravity_rad(self) -> float:
        """Returns the current angle of gravity, in radians. Use ``set_angleofgravity_rad`` to set it manually."""
        return self._a_rad
    def set_angleofgravity_rad(self, gravity_angle_rad:float) -> None:
        """Manually sets the angle of gravity, in radians, to the given value."""
        self._a_rad = gravity_angle_rad
    def get_angleofgravity_deg(self) -> float:
        """Returns the current angle of gravity, in degrees. Use ``set_angleofgravity_deg`` to set it manually."""
        return None if self._a_rad is None else self._a_rad * self.DEG_PER_RAD
    def set_angleofgravity_deg(self, gravity_angle_deg:float) -> None:
        """Manually sets the angle of gravity, in degrees, to the given value."""
        self._a_rad = None if gravity_angle_deg is None else gravity_angle_deg * self.RAD_PER_DEG

    ###### TIP POSITION SHORT VERSION
    def set_x_px(self, x_tip_position_px:float) -> None:
        self._x_px = x_tip_position_px
    def get_x_px(self) -> float:
        return self._x_px
    def set_y_px(self, y_tip_position_px:float) -> None:
        self._y_px = y_tip_position_px
    def get_y_px(self) -> float:
        return self._y_px
    def set_xy_px(self, xy_tip_position_px:float) -> None:
        self._x_px, self._y_px = xy_tip_position_px
    def get_xy_px(self) -> Tuple[float, float]:
        return (self._x_px, self._y_px)

    ### TIP POSITION LONG VERSION
    def set_xtippos_px(self, x_tip_position_px:float) -> None:
        self._x_px = x_tip_position_px
    def get_xtippos_px(self) -> float:
        return self._x_px
    def set_ytippos_px(self, y_tip_position_px:float) -> None:
        self._y_px = y_tip_position_px
    def get_ytippos_px(self) -> float:
        """Get the y position of the tip, in pixels. Use ``set_xytippos_px`` to set it."""
        return self._y_px
    def set_xytippos_px(self, xy_tip_position_px:float) -> None:
        self._x_px, self._y_px = xy_tip_position_px
    def get_xytippos_px(self) -> Tuple[float, float]:
        """Get the (x, y) position of the tip, in pixels. Use ``set_xytippos_px`` to set it."""
        return (self._x_px, self._y_px)

    ### TIP RADIUS OF CURVATURE SHORT VERSION
    def get_r_px(self) -> float:
        return self._r_px
    def set_r_px(self, r0_px:float) -> None:
        self._r_px = r0_px
    def get_r_mm(self) -> float:
        return None if (self._px_per_mm is None or self._r_px is None) else self._r_px / self._px_per_mm
    def set_r_mm(self, r0_mm:float) -> None:
        self._r_px = None if (r0_mm is None or self._px_per_mm is None) else r0_mm * self._px_per_mm

    ### TIP RADIUS OF CURVATURE LONG VERSION
    def get_tipradius_px(self) -> float:
        """
        Returns the radius of curvature at the tip of the drop, in pixels.

        Use :meth:`describe <pypendentdrop.Parameters.set_tipradius_px>` to set this value manually.

        Returns
        -------
        r0_px : float
            The tip radius, in pixels.
        """
        return self._r_px
    def set_tipradius_px(self, r0_px:float) -> None:
        """
        Manually sets the radius of curvature at the tip of the drop, in pixels.

        Parameters
        ----------
        r0_px : float
            The tip radius, in pixels.
        """
        self._r_px = r0_px
    def get_tipradius_mm(self) -> float:
        """Get the radius of curvature at the tip, in millimeters. Use ``set_tipradius_mm`` to set it."""
        return None if (self._px_per_mm is None or self._r_px is None) else self._r_px / self._px_per_mm
    def set_tipradius_mm(self, r0_mm:float) -> None:
        self._r_px = None if (r0_mm is None or self._px_per_mm is None) else r0_mm * self._px_per_mm

    ### CAPILLARY LENGTH SHORT VERSION
    def get_l_px(self) -> float:
        return self._l_px
    def set_l_px(self, lcap_px:float) -> None:
        self._l_px = lcap_px
    def get_l_mm(self) -> float:
        return None if (self._px_per_mm is None or self._l_px is None) else self._l_px / self._px_per_mm
    def set_l_mm(self, lcap_mm:float) -> None:
        self._l_px = None if (lcap_mm is None or self._px_per_mm is None) else lcap_mm * self._px_per_mm

    ### CAPILLARY LENGTH LONG VERSION
    def get_caplength_mm(self) -> float:
        """
        Returns the capillary length, in millimeters.

        Use :meth:`describe <pypendentdrop.Parameters.set_caplength_mm>` to set this value manually.

        Returns
        -------
        lcap_mm : float
            The capillary length, in millimeters.
        """
        return self.get_l_mm()
    def set_caplength_mm(self, lcap_mm:float):
        """
        Manually sets the capillary length, in millimeters.

        Parameters
        ----------
        lcap_mm : float
            The capillary length, in millimeters.
        """
        self.set_l_mm(lcap_mm=lcap_mm)
    def get_caplength_px(self) -> float:
        """
        Returns the capillary length, in pixels.

        Use :meth:`describe <pypendentdrop.Parameters.set_caplength_px>` to set this value manually.

        Returns
        -------
        lcap_px : float
            The capillary length, in pixels.
        """
        return self.get_l_px()
    def set_caplength_px(self, lcap_px:float) -> None:
        """
        Manually sets the capillary length, in pixels.

        Parameters
        ----------
        lcap_px : float
            The capillary length, in pixels.
        """
        self.set_l_px(lcap_px=lcap_px)

    ### GRAVITY ACCELERATION SHORT VERSION
    def set_g(self, g:Optional[float]) -> None:
        self._g = g
    def get_g(self):
        return self._g

    ### GRAVITY ACCELERATION LONG VERSION
    def set_gravityacc(self, g:Optional[float]) -> None:
        """Sets the value for the acceleration of gravity for this experiment.

        Parameters
        ----------
        g : float
            The acceleration of gravity, in m/s^2
        """
        self._g = g
    def get_gravityacc(self):
        """Returns the acceleration of gravity for this experiment.

        Use :meth:`describe <pypendentdrop.Parameters.set_gravityacc>` to modify this value.

        Returns
        -------
        g : float
            The acceleration of gravity, in m/s^2.
            The default value is 9.81.
        """
        return self._g

    ### DENSITY CONTRAST SHORT VERSION
    def set_d(self, densitycontrast_kgL:Optional[float]) -> None:
        self._d = densitycontrast_kgL
    def get_d(self):
        return self._d

    ### DENSITY CONTRAST LONG VERSION
    def set_densitycontrast(self, densitycontrast_kgL:Optional[float]) -> None:
        self._d = densitycontrast_kgL
    def get_densitycontrast(self):
        """Get the density contrast between the fluids, in kg/L (typically 1 for water/air). Use ``set_densitycontrast`` to set it."""
        return self._d
    def get_dimensionlessTipRadius(self):
        return None if (self._r_px is None or self._l_px is None) else self._r_px / self._l_px

    def get_fitparams(self) -> Fitparams:
        """Returns the parameters in a ``scipy.minimize``-friendly way."""
        return [self.get_a_rad(), self.get_x_px(), self.get_y_px(), self.get_r_px(), self.get_l_px()]

    def can_show_tip_position(self) -> bool:
        x_is_ok = not( (self.get_x_px() or 0) == 0 )
        y_is_ok = not( (self.get_y_px() or 0) == 0 )
        return x_is_ok * y_is_ok
    def can_estimate(self) -> bool:
        """Decides if an auto-estimation of the parameters is possible."""
        return not( (self.get_px_density() or 0) == 0 )
    def can_optimize(self) -> bool:
        """Decides if an optimization of the parameters is possible."""
        r0_is_ok = not( (self.get_r_px() or 0) == 0 )
        lcap_is_ok = not( (self.get_l_px() or 0) == 0 )
        px_density_is_ok = not( (self.get_px_density() or 0) == 0 )
        return r0_is_ok * lcap_is_ok * px_density_is_ok

    def get_bond(self):
        """Returns the Bond number (square of the (tip radius)/(capillary length) ratio)"""
        return None if (self._r_px is None or self._l_px is None) else (self._r_px / self._l_px)**2

    def get_surface_tension_mN(self):
        """Returns the surface tension in mN (or None if it cannot be computed)."""
        return None if ((self.get_g() is None) or (self.get_d() is None) or (self.get_l_mm() is None)) else (self.get_d() * self.get_g() * self.get_l_mm()**2)


# paremeters
def image_centre(image:np.ndarray) -> np.ndarray:
    """
    Returns the (x, y) coordinates of the center of an image (around which it is pertinent to rotate the contour).

    :param image:
    :return:
    """
    return np.array(image.shape[1::-1]) / 2

def getrotationandscalematrix(centre, angle:float=0., scalefactor:float=1.):
    """
    Gets the matrix allowing the rotation and scaling of a contour (or an image) around a point.
    Shamelessly inspired from OpenCV's cv2.getRotationMatrix2D.

    :param centre:
    :param angle:
    :param scalefactor:
    :return:
    """
    c = math.cos(angle) * scalefactor
    s = math.sin(angle) * scalefactor
    return np.array([[c, s, (1-c)*centre[0] - s*centre[1]],
                     [-s, c, s*centre[0] + (1-c)*centre[1]]])

def rotate_and_scale(contour, angle:float=0., centre:Optional[Union[Tuple, List, np.ndarray]]=None, scalefactor=1.):
    """
    Rotates and/or scales a contour around a center **centre**.

    :param contour: shape (2, N)
    :param angle: angle, in radians
    :param centre: coordinates (x, y), in px
    :param scalefactor:
    :return:
    """
    if centre is None: # centre of rotation
        centre = (0, 0)
    rot_mat = getrotationandscalematrix(centre, angle=angle, scalefactor=scalefactor)
    x_rotated = rot_mat[0,0] * contour[0] + rot_mat[0, 1] * contour[1] + rot_mat[0, 2]
    y_rotated = rot_mat[1,0] * contour[0] + rot_mat[1, 1] * contour[1] + rot_mat[1, 2]
    return x_rotated, y_rotated

def estimate_parameters(image:np.ndarray, contour:np.ndarray, px_per_mm) -> Parameters:
    """Provides a coarse estimation of the parameters using (hopefully) robust techniques.

    Parameters
    ----------
    image : ndarray
        The image of the drop
    contour : ndarray
        The contour of the drop, with shape (2, N)
    px_per_mm : float
        The pixel density of the image.

    Returns
    -------
    parameters: Parameters
        An estimation of the parameters describing the drop.

    """
    if image.shape == (2,): # for old code compatibility # depreciate #todo remove this one day
        centre_of_image = image
    else:
        centre_of_image = image_centre(image)

    ### ANGLE OF GRAVITY
    gravity_angle:float = 0.
    try:
        slope, intercept = np.polyfit(contour[1], contour[0], deg=1)
        abc0 = np.array([slope, -1., intercept])
        # abc0 = np.array([0., -1., contour[0].mean()])

        def dist(abc, contour:np.ndarray) -> float:
            return np.sum((abc[0]*contour[1] + abc[1]*contour[0] +abc[2])**2/(abc[0]**2+abc[1]**2))


        bestline = minimize(dist, abc0, args=(contour))
        a, b, c = bestline.x


        gravity_angle = np.arctan(-a/b)
        # we do a trick to have the angle between -90 and 90 deg
        gravity_angle = (np.pi/2 + np.arctan2(-a, b))%(np.pi) - np.pi/2

        if (np.abs(gravity_angle*180/np.pi) > 15):
            warning(f'WARN: the angle of gravity was detected to {round(gravity_angle*180/np.pi, 2)} deg.')
            warning(f'WARN: This is likely an error so I put it back to {gravity_angle}.')
            gravity_angle = 0
    except:
        warning(f"WARN: couldn't get gravity angle. Falling back to {gravity_angle}")

    trace(f"\tFound gravity_angle={gravity_angle*180/np.pi} deg")

    # Now we need to rotate the contour in order to correctly estimate the other parameters
    contour_tiltcorrected = rotate_and_scale(contour, angle=-gravity_angle, centre=centre_of_image)

    ### Position of the tip

    # we guess it from the tilt-corrected contour
    y_tip_position_tiltcorrected = contour_tiltcorrected[1].max()
    x_tip_position_tiltcorrected = np.mean(contour_tiltcorrected[0][np.argmax(contour_tiltcorrected[1])])

    # we translate that to the real contour
    x_tip_position, y_tip_position = rotate_and_scale([x_tip_position_tiltcorrected, y_tip_position_tiltcorrected],
                                                      angle=gravity_angle, centre=centre_of_image)

    trace(f"\tFound x_tip_position={x_tip_position} px")
    trace(f"\tFound y_tip_position={y_tip_position} px")

    ### Radius of curvature
    De = contour_tiltcorrected[0].max() - contour_tiltcorrected[0].min()

    right = contour_tiltcorrected[0] > x_tip_position_tiltcorrected
    left = np.bitwise_not(right)
    Ds_height = y_tip_position_tiltcorrected - De
    x_ds_r = contour_tiltcorrected[0][right][np.argmin((contour_tiltcorrected[1][right] - Ds_height)**2)]
    x_ds_l = contour_tiltcorrected[0][left][np.argmin((contour_tiltcorrected[1][left] - Ds_height)**2)]
    Ds = x_ds_r - x_ds_l

    # Method 1: coarse estimate : max radius of the drop
    r0_px_coarse = De / 2
    # Method 2 : using the ugly formulas
    sigma = De/Ds
    bond = 0.12836 - 0.7577*sigma + 1.7713*sigma**2 - 0.5426*sigma**3
    ratio = 0.9987 + 0.1971 * bond - 0.0734 * bond**2 + 0.34708*bond**3
    r0_px_fine = De / (2 * ratio)

    # trace(f'De: {De} | Ds = {Ds}')
    # trace(f'sigma: {sigma} | bond2 = {bond}')

    # Method 3 : using a fit
    r0_px_fit = r0_px_coarse
    try:
        # finer estimation : we take the points near the tip (at a distance < r0_estimation_radius, in px) and do a 2nd order polynomial fitparams_handpicked
        r0_estimation_radius = r0_px_coarse/4 # this is tipNeighbourhood is pendent drop
        nearthetip = (contour_tiltcorrected[0] - x_tip_position_tiltcorrected)**2 + (contour_tiltcorrected[1] - y_tip_position_tiltcorrected)**2 < r0_estimation_radius**2
        x_contour_neartip = contour_tiltcorrected[0][nearthetip]
        y_contour_neartip  = contour_tiltcorrected[1][nearthetip]
        curvature_px = np.polyfit(x_contour_neartip, y_contour_neartip, 2)[0]
        r0_px_fit = np.abs(1/curvature_px)/2
    except:
        debug('Fit failed')

    # trace(f"\tr0= {round(r0_px_coarse, 2)} px [coarse] / {round(r0_px_fine, 2)} px [fine] / {round(r0_px_fit, 2)} px [fit] ")
    trace(f"\tr0= {round(r0_px_coarse/px_per_mm, 2)} mm [coarse] / {round(r0_px_fine/px_per_mm, 2)} mm [fine] / {round(r0_px_fit/px_per_mm, 2)} mm [fit] ")

    # lcap_px_coarse = np.sqrt(r0_px_coarse**2 / bond)
    # lcap_px_fine = np.sqrt(r0_px_fine**2 / bond)
    # lcap_px_fit = np.sqrt(r0_px_fit**2 / bond)
    # trace(f"\tlcap= {round(lcap_px_coarse/px_per_mm, 2)} mm [coarse] / {round(lcap_px_fine/px_per_mm, 2)} mm [fine] / {round(lcap_px_fit/px_per_mm, 2)} mm [fit] ")

    lcap_px_clueless:float = 2.7*px_per_mm # clueless guess : the fluid is pure water
    lcap_px = lcap_px_clueless

    params_estimated = Parameters()
    params_estimated.set_px_density(px_per_mm)
    params_estimated.set_a_rad(gravity_angle)
    params_estimated.set_x_px(x_tip_position)
    params_estimated.set_y_px(y_tip_position)
    params_estimated.set_r_px(r0_px_fit)
    params_estimated.set_l_px(lcap_px)

    params_estimated.describe(printfn=trace, descriptor='(estimated)')
    debug(f'Difference between contour and estimated profile: {round(compute_gap_pixel(contour, params_estimated), 2)} px^2')

    return params_estimated

### profile computing

def deriv(s_, rpkz):
    """
    The derivative along the arc of the variables.

    :param s_: (unused) The integration coordinate, the dimensionless arclength s
    :param rpkz: 4-array : [R, Psi, Kappa=dpsi/ds, Z]
    :return:
    """
    c = math.cos(rpkz[1])
    s = math.sin(rpkz[1])
    oneoverr = 1/rpkz[0]
    return [c, rpkz[2], -s + c * (s*oneoverr - rpkz[2])*oneoverr, s]

def greater_possible_zMax(tipRadius:float) -> float:
    """
    The max value zmax can take, from Daerr, pendent drop.

    :param tipRadius:
    :return:
    """
    return min(3.5, 3 / tipRadius) if tipRadius > .5 else 2 * tipRadius * (1 + 1.55*tipRadius)

def compute_nondimensional_profile(tipRadius:float, ds:float = 1e-3, approxLimit:float = 0.2, zMax:float = None) -> Tuple[np.ndarray, np.ndarray]:
    """Computes the nondimensional profile, starting at the bottom and up to an height zMax.

    About the default value of ds:
    The length in px of the contour is L. We want around 2 points per pixel, and the length of the dimensionless drop will be
    s = L / (px_per_mm * capillarylength_in_mm).
    Thus taking a typical 100 px/mm (microscopic resolution is 1000) and a capillary length of 2.5, we have
    ds = s / (2 * L) = 1 / (2 * 100 * 2.5) = 2e-3
    To be safe, we take ds = 1e-3 by default.
    If using high resolution camera + excellent optics + optimal conditions, one might want it smaller.

    About the solver : We use LSODA form LAPACK (scipy.integrate's odeint). Good ol' Fortran never dissapoints.

    Parameters
    ----------
    tipRadius : float, optional
        The dimensionless tip radius r_0/l_c (square root of the Bond number).
    ds : float, optional
        The dimensionless integration step. The default value (1e-3) should be enough for most applications.
    approxLimit : float, optional
        The fraction of capillary length or tip radius (whichever is smaller) up to which we use the approximate solution.
        todo: compute the difference between "true" profile (approxLimit = 1e-3) and the others to see what is the maximum reasonable approxlimit.
    zMax : float, optional
        The maximum height of the drop [useless]

    Returns
    -------
    R, Z : Tuple[ndarray]
        The right (R > 0) half of the dimensionless profile of the drop.

    """
    if zMax is None:
        zMax = greater_possible_zMax(tipRadius)

    ### NEAR THE TIP (singularity handling by asymptotic development)

    # curvilinear coordinate where we switch from the approximate solution to integration
    sLimit = approxLimit * min(1., tipRadius) # sLimit << (Cap length, Tip radisu)

    # the curvilinear coordinates we will use
    s_neartip = np.arange(0, sLimit, ds, dtype=float)

    RPKZ_neartip = np.empty((len(s_neartip), 4))

    # use the approximation near the tip
    RPKZ_neartip[:, 0] = tipRadius * np.sin(s_neartip/tipRadius) + s_neartip**5/(40*tipRadius**2)
    RPKZ_neartip[:, 1] = s_neartip * (1 - s_neartip**2 / 8) / tipRadius
    RPKZ_neartip[:, 2] = (1 - 3 * s_neartip**2 / 8) / tipRadius
    RPKZ_neartip[:, 3] = s_neartip**2 * ( 1 + (1/16 - 1/(12*tipRadius**2)) * s_neartip**2 ) / (2*tipRadius)

    # AWAY FROM THE TIP
    N_awayfromtip = int(5 / ds)+1
    s_awayfromtip = np.arange(N_awayfromtip) * ds + s_neartip[-1]

    RPKZ_initial = RPKZ_neartip[-1, :]

    RPKZ_awayfromtip:np.ndarray = odeint(deriv, RPKZ_initial, s_awayfromtip, tfirst=True)
    # # if anyone wants to use RK45 explicitely (works but slower because the new and shiny solve_ivp is slower than the good ol' odeint):
    # RPKZ_awayfromtip = solve_ivp(deriv, [s_awayfromtip[0], s_awayfromtip[-1]], RPKZ_initial, method='RK45', t_eval=s_awayfromtip)
    # RPKZ_awayfromtip = RPKZ_awayfromtip.T

    Nmax_zMax_attained = N_awayfromtip if RPKZ_awayfromtip[-1, 3] <= zMax else np.argmax(RPKZ_awayfromtip[:, 3] > zMax)
    Nmax_axisymmetry = N_awayfromtip if RPKZ_awayfromtip[:, 0].min() > 0 else np.argmax(RPKZ_awayfromtip[:, 0] <= 0)
    Nmax_goingup = N_awayfromtip if RPKZ_awayfromtip[:, 1].min() > 0 else np.argmax(RPKZ_awayfromtip[:, 1] <= 0)
    Nmax_neck = np.argmin(np.maximum( RPKZ_awayfromtip[1:,1] > np.pi/2 , RPKZ_awayfromtip[:-1,1] <= np.pi/2))
    if Nmax_neck == 0: Nmax_neck = N_awayfromtip

    RPKZ_awayfromtip = RPKZ_awayfromtip[:min(Nmax_zMax_attained, Nmax_axisymmetry, Nmax_goingup, Nmax_neck),:]

    # combine everything
    RPKZ = np.concatenate((RPKZ_neartip, RPKZ_awayfromtip), axis=0)

    return RPKZ[:, 0], RPKZ[:, 3]

def integrated_contour(parameters:Parameters) -> np.ndarray:
    """The computed profile in pixel coordinates.

    Parameters
    ----------
    parameters : Parameters

    Returns
    -------
    contour: Tuple[ndarray, ndarray]
        The computed profile in pixel coordinates (shape 2, N).

    """

    tipRadius = parameters.get_dimensionlessTipRadius()
    x_tip_position, y_tip_position = parameters.get_xy_px()
    gravity_angle = parameters.get_a_rad()
    caplength_px = parameters.get_l_px()

    # hence the profile
    R, Z = compute_nondimensional_profile(tipRadius)

    # FOR DISPLAY : WE MAKE Rdim AND Zdim, TO SHOW ON IMAGE

    # we scale, then rotate (these two can be exchanged), then move
    Rdim, Zdim = R.copy(), Z.copy()

    # complete, using axisymmetry
    Rdim = np.concatenate((-Rdim[::-1], Rdim))
    Zdim = np.concatenate(( Zdim[::-1], Zdim))

    # Reverse Z because we have this weird image coordinated system
    Zdim *= -1
    # scaling (first by capillary length to dimensionalize, then by pizel size to fit the image)
    # rotate, around the tip at (0,0)
    Rdim, Zdim = rotate_and_scale([Rdim, Zdim], angle=gravity_angle, scalefactor=caplength_px)

    # moving
    Rdim = Rdim + x_tip_position
    Zdim = Zdim + y_tip_position

    return np.array((Rdim, Zdim))

#compare computed profile to real profile
def compute_gap_dimensionless(contour:np.ndarray, parameters:Parameters) -> float:
    """The dimensionless area between the detected contour and the computed profile.

    Parameters
    ----------
    contour : ndarray
        The pixel coordinates of the detected contour to be compared with the computed profile, shape=(2, N)
    parameters : Parameters
        The parameters to compute the theoretical profile.

    Returns
    -------
    gap_dimensionless : float

    """
    fitparams:Fitparams = parameters.get_fitparams()

    return compute_gap_dimensionless_fromfitparams(fitparams=fitparams, contour=contour)
def compute_gap_pixel(contour:np.ndarray, parameters:Parameters) -> float:
    """The area between the detected contour and the computed profile, in squared pixels.

    Parameters
    ----------
    contour : ndarray
        The pixel coordinates of the detected contour to be compared with the computed profile, shape=(2, N)
    parameters : Parameters
        The parameters to compute the theoretical profile.

    Returns
    -------
    gap_in_squared_px : float

    """
    l_px = parameters.get_l_px()

    return compute_gap_dimensionless(contour, parameters)**(2/3) * l_px**2
def compute_gap_dimensionless_fromfitparams(fitparams:Fitparams, contour) -> float:
    gravity_angle, x_tip_position, y_tip_position, r0_px, capillary_length_px = fitparams

    tipRadius = r0_px / capillary_length_px

    # hence the profile
    R, Z = compute_nondimensional_profile(tipRadius)

    # FOR COMPUTE THE DIFF : we take it backward
    XY = contour.copy()

    #moving
    XY[0] -= x_tip_position
    XY[1] -= y_tip_position

    #  rotating and scaling
    XY = rotate_and_scale(XY, angle=-gravity_angle, scalefactor=-1 / capillary_length_px)

    # # cutting off :
    # XY = np.take(XY, np.where(XY[1] < Z.max())[0], axis=1)


    # separating the two sides
    rightside = XY[0] > 0
    X1, Y1 = np.take(XY, np.where(rightside)[0], axis=1)
    # X2, Y2 = -X[X < 0], Y[X < 0]
    # X2, Y2 = XY[:, np.bitwise_not(rightside)
    X2, Y2 = np.take(XY, np.where(np.bitwise_not(rightside))[0], axis=1)
    X2 *= -1

    # the differences
    R1 = np.interp(Y1, Z, R) # the radius corresponding to the side 1
    R2 = np.interp(Y2, Z, R) # the radius corresponding to the side 2

    R1[Y1 > Z.max()] = R[np.argmax(Z)]
    R2[Y2 > Z.max()] = R[np.argmax(Z)]

    R1[Y1 < Z.min()] *= 0
    R2[Y2 < Z.min()] *= 0
    DX1 = X1 - R1
    DX2 = X2 - R2

    difference = np.abs(trapezoid(DX1**2, Y1)) + np.abs(trapezoid(DX2**2, Y2))
    return difference

# compute the optimal profile
def optimize_profile(contour:np.ndarray, parameters_initialguess:Parameters,
                     to_fit:Optional[List[bool]]=None,
                     maxiter:Optional[int]=None, method:Optional[str]=None) -> Tuple[bool, Parameters]:
    """Computes the optimized parameters for the given ``contour``.

    Minimises the :func:`dimensionless gap <pypendentdrop.compute_gap_dimensionless>` between the contour and the computed profile,
    using scipy's ``minimize``.

    Parameters
    ----------
    contour : ndarray
        The pixel coordinates of the detected contour to be compared with the computed profile, shape=(2, N)
    parameters_initialguess : Parameters
        The initial guess of the parameters
    to_fit : Tuple[bool, bool, bool, bool, bool], optional
        Whether of not to fit the parameters todo: better document this (SEE EXAMPLES)
    maxiter : int, optional
        The maximum number of iterations. Depends on the method !
    method : str, optional
        The method passed to ``minimize``. Must be a bound-constrained method. Nelder-Mead works quite well.

    Returns
    -------

    """
    if method is None:
        method = 'Nelder-Mead'
    if to_fit is None:
        to_fit = [True, True, True, True, True]
    if maxiter is None:
        if method == 'Powell':
            maxiter = 100
        elif method == 'Nelder-Mead':
            maxiter = 2000 # should not take long

    default_bounds = [(None, None),  # gravity_angle
                      (None, None),  # y_tip_position
                      (None, None),  # x_tip_position
                      (0, None),  # r0_mm
                      (0, None)] # capillary length (mm)

    fitparams_initial = parameters_initialguess.get_fitparams()

    bounds = [default_bounds[i] if to_fit[i] else (fitparams_initial[i], fitparams_initial[i]) for i in range(len(fitparams_initial))]

    # we remove the top 5 pixels, i.e. the points too close to the top edge
    # gravity_angle = parameters_initialguess.get_a_rad()
    nottooclosefromthetop = contour[1]-contour[1].min() > 5/np.cos(parameters_initialguess.get_a_rad())
    contour_opti = contour[:, nottooclosefromthetop]

    options={}
    if maxiter is not None:
        options['maxiter'] = maxiter

    # pendent drop uses Powell's method with maxiter = 100 (typically maxiter = 10),
    # but I found the Nelder-Mead algorithm to be 25% faster ?
    # for method = Nelder-Mead, options={'adaptive':False, 'disp':False}
    trace(f'Fitparams initial: {[round(fpi, 5) for fpi in fitparams_initial]}')
    trace(f'Bounds: {bounds}')
    trace(f'Contour shape: {contour.shape} (min-max = {contour.min()}-{contour.max()})')
    trace(f'Optimization: {method} method (options: {options})')

    t1 = time.time()
    minimization = minimize(compute_gap_dimensionless_fromfitparams, x0=np.array(fitparams_initial), args=(contour_opti),
                            bounds=bounds, method=method, options=options)
    t2 = time.time()
    debug(f'optimize_profile: Optimisation time: {int((t2-t1)*1000)} ms')

    optimization_success = minimization.success

    debug(f'optimize_profile: {minimization.nit} iterations, {minimization.nfev} calls to function')
    trace(f'optimize_profile: Minimization message: {minimization.message}')

    if not(minimization.success):
        info('Minimizaton failed')
        debug(f'Minimization unsuccessful: {minimization.message}')
        return False, parameters_initialguess


    # copy `by hand'
    # todo implement a copy method for the Parameters class and use it here
    parameters_opti = Parameters()
    parameters_opti.set_px_density(parameters_initialguess.get_px_density())
    parameters_opti.set_g(parameters_initialguess.get_g())
    parameters_opti.set_d(parameters_initialguess.get_d())
    # update with the optimized values
    parameters_opti.set_a_rad(minimization.x[0])
    parameters_opti.set_x_px(minimization.x[1])
    parameters_opti.set_y_px(minimization.x[2])
    parameters_opti.set_r_px(minimization.x[3])
    parameters_opti.set_l_px(minimization.x[4])

    parameters_opti.describe(printfn=trace, descriptor='(optimized)')
    debug(f'Difference between contour and optimized profile: {round(compute_gap_pixel(contour, parameters_opti), 2)} px^2')

    return True, parameters_opti



from typing import Tuple, Union, Optional, Dict, Any, List
import numpy as np
from PIL import Image

from .. import error, warning, info, debug, trace


# import image and threshold
def import_image(file_path:Optional[str] = None) -> Tuple[bool, np.ndarray]:
    """Imports an image in grayscale.

    Tries to import the image at `file_path`.
    Returns True and the image in grayscale if the import succeeded.
    Returns False and a random 128 x 128 matrix with values 0-255 if the import failed.

    Parameters
    ----------
    file_path : str
        The path to the file to open. Can be None.

    Returns
    -------
    success : bool
        Whether of not the image could be imported.

    data : ndarray
        The image in grayscale.

    """
    success = False
    data = None
    if file_path is None:
        debug('import_image: File path provided is None. Failing to import')
    else:
        try:
            imagedata = Image.open(file_path)
            data = np.asarray(imagedata, dtype="float")
            if len(data.shape) > 2: # go to gray
                data = np.mean(data, axis=2)
            success = True
        except:
            # todo: add precise reason why it did not work
            warning(f'pypendentdrop.import_image: Could not import the image at path "{file_path}"')
    if not success:
        data = np.random.randint(0, 255, (128, 128))
    trace(f'import_image: Imported image {file_path} with success: {success}')
    return success, data



import sys
import argparse

import pypendentdrop as ppd
from pypendentdrop import logfacility
from pypendentdrop.ppd_parse import generate_parser
from pypendentdrop import ppdplot

testdata_filepath = 'tests/testdata/water_2_rotated.tif'
testdata_pxldensity = 57.0
testdata_g = 9.81
testdata_d = 1.00
testdata_roi = None

parser = generate_parser('test-water2_rotated')
args = parser.parse_args(['-n', str(testdata_filepath), '-p', str(testdata_pxldensity),
                          '-g', str(testdata_g), '-d', str(testdata_d),
                          '-vvv', '-o', 'water2_rotated'])

if __name__ == "__main__":
    logfacility.set_verbose(args.v)

    imagefile = args.n
    if imagefile is None:
        ppd.logger.error(f'No image file provided.')
        ppd.logger.error(f'Use -n to specify the image you want to analyze (e.g. -n {testdata_filepath})')
        sys.exit(101)

    ppd.logger.debug(f'Image path provided: {imagefile}')

    px_per_mm = args.p
    if px_per_mm is None:
        ppd.logger.error(f'No pixel density provided.')
        ppd.logger.error(f'Use -p to specify the pixel density, in mm/px (e.g. -p {testdata_pxldensity})')
        sys.exit(102)

    ppd.logger.debug(f'Pixel density provided: {px_per_mm} px/mm')

    import_success, img = ppd.import_image(imagefile)

    if import_success:
        ppd.logger.debug(f'Import image successful.')
    else:
        ppd.logger.error(f'Could not retreive the image at {imagefile}')
        sys.exit(200)

    height, width = img.shape
    ppd.logger.debug(f'Image shape: {height}x{width}')

    roi = ppd.format_roi(img, [args.tlx, args.tly, args.brx, args.bry])
    ppd.logger.debug(f'roi = {roi}')

    threshold = args.t
    if threshold is None:
        ppd.logger.debug('Threshold not provided, using auto_threshold to provide it.')
        threshold = ppd.auto_threshold(img, roi=roi)

    ppd.logger.debug(f'Threshold level: {threshold}')

    lines = ppd.detect_contourlines(img, threshold, roi=roi)
    linelengths = [len(line) for line in lines]

    ppd.logger.debug(f'Number of lines: {len(lines)}, lengths: {linelengths}')

    cnt = ppd.detect_main_contour(img, threshold, roi=roi)

    ppd.logger.debug(f'Drop contour: {cnt.shape[1]} points')

    estimated_parameters = ppd.estimate_parameters(img, cnt, px_per_mm)

    args_parameters = ppd.Parameters()
    args_parameters.set_px_density(px_per_mm)
    args_parameters.set_a_deg(args.ai)
    args_parameters.set_x_px(args.xi)
    args_parameters.set_y_px(args.yi)
    args_parameters.set_r_mm(args.ri)
    args_parameters.set_l_mm(args.li)
    args_parameters.describe(printfn=ppd.trace, descriptor='from arguments')

    initial_parameters = ppd.Parameters()
    initial_parameters.set_px_density(px_per_mm)
    initial_parameters.set_a_deg(args.ai or estimated_parameters.get_a_deg())
    initial_parameters.set_x_px(args.xi or estimated_parameters.get_x_px())
    initial_parameters.set_y_px(args.yi or estimated_parameters.get_y_px())
    initial_parameters.set_r_mm(args.ri or estimated_parameters.get_r_mm())
    initial_parameters.set_l_mm(args.li or estimated_parameters.get_l_mm())
    initial_parameters.describe(printfn=ppd.debug, descriptor='initial')

    ppd.logger.debug(f'chi2: {ppd.compute_gap_dimensionless(cnt, parameters=initial_parameters)}')

    to_fit = [args.af, args.xf, args.yf, args.rf, args.lf]

    ppd.logger.debug(f'to_fit: {to_fit}')

    opti_success, opti_params = ppd.optimize_profile(cnt, parameters_initialguess=initial_parameters, to_fit=to_fit,
                                                     method=None)

    if opti_success:
        opti_params.describe(printfn=ppd.info, descriptor='optimized')

        ppd.logger.debug(f'chi2: {ppd.compute_gap_dimensionless(cnt, parameters=opti_params)}')
    else:
        ppd.logger.warning('Optimization failed :( Falling back to the estimated parameters.')

    # r0_mm = opti_params[3]
    # caplength_mm = opti_params[4]
    #
    # bond = (r0_mm / caplength_mm)**2
    #
    # print(f'Bond number: {round(bond, 3)}')

    # rhog = args.g
    # if rhog is None:
    #     ppd.logger.error(f'No density contrast provided, could not compute surface tension.')
    #     ppd.logger.error(f'Use -g to specify the density contrast times gravity (e.g. -g {testdata_rhog})')
    # else:
    #     gamma = rhog * caplength_mm**2
    #     print(f'Surface tension gamma: {round(gamma, 3)} mN/m')

    if args.o is not None:
        import matplotlib
        matplotlib.use('Qt5Agg')
        import matplotlib.pyplot as plt

        ppdplot.generate_figure(img, cnt, initial_parameters,
                             prefix=args.o, comment='estimated parameters', suffix='_initialestimate', filetype='pdf', roi=roi)
        if opti_success:
            ppdplot.generate_figure(img, cnt, opti_params,
                                 prefix=args.o, comment='optimized parameters', suffix='_optimalestimate', filetype='pdf', roi=roi)
        plt.show()

    sys.exit(0)



#!/usr/bin/env python

import pypendentdrop as ppd

# the path to the image file
filepath = './tests/testdata/water_2.tif'
# The Region Of Interest: position of Top Left (TL) and Bottom Right (BR) corners of the ROI
# roi = [TLx, TLy, BRx, BRy]
# Use roi = None or roi = [None, None, None, None] if you do not care
roi = [10, 90, 300, 335] # This is reasonable for the file water_2.tif
roi = None # this also works but is not recommended

# 1. Import the image
import_successful, image = ppd.import_image(filepath)

if not import_successful:
    raise FileNotFoundError(f'Could not import image at {filepath}')

# 2. Choose a threshold (here automatically)
threshold = ppd.auto_threshold(image, roi=roi)

# 3. Detect the contour
contour = ppd.detect_main_contour(image, threshold, roi=roi)

# 4. Pixel density of the image, in px/mm
pxldensity = 57.0

# 5. Estimate roughly the parameters of the drop
estimated_parameters = ppd.estimate_parameters(image, contour, pxldensity)

# Set manually an estimation of the capillary length
estimated_parameters.set_caplength_mm(2.65)

# Print the estimated parameters in the console
estimated_parameters.describe(descriptor='estimated')

# 6. Optimize these parameters
optimization_successful, optimized_parameters = ppd.optimize_profile(contour, estimated_parameters)

if optimization_successful:
    # Print the optimized parameters in the console
    optimized_parameters.describe(descriptor='optimized')

    # Print the bond number corresponding to the drop
    print(f'Bond number: {round(optimized_parameters.get_bond(), 3)}')

    # The density contrast and gravity
    deltarho = 1.00
    g = 9.81
    optimized_parameters.set_densitycontrast(deltarho)
    optimized_parameters.set_gravityacc(g)

    # 7. Compute the surface tension
    print(f'Surface tension gamma: {round(optimized_parameters.get_surface_tension_mN(), 3)} mN/m')
else:
    print('Optimization failed :(')

# Let's plot some stuff
import matplotlib.pyplot as plt
from pypendentdrop import ppdplot # a small subpackage to help at plotting

if not optimization_successful:
    fig, (ax1, ax1) = plt.subplots(1, 1)
    ppdplot.plot_image_contour(ax1, image, contour, estimated_parameters, 'estimated', roi=roi)
    plt.savefig('deleteme_estimatedparameters.png', dpi=300)
else:
    ### Plotting a comparison between the estimated and optimized parameters

    fig, (ax1, ax2) = plt.subplots(1, 2)
    ppdplot.plot_image_contour(ax1, image, contour, estimated_parameters, 'estimated', roi=roi)
    ppdplot.plot_image_contour(ax2, image, contour, optimized_parameters, 'optimized', roi=roi)
    plt.savefig('deleteme_comparison.png', dpi=300)



from typing import Tuple, Union, Optional, Dict, Any, List
import numpy as np
import warnings
from contourpy import contour_generator, LineType

from .. import error, warning, info, debug, trace

# Region OF Interest management
Roi = Optional[List[Optional[int]]]


def format_roi(image: np.ndarray, roi: Roi = None) -> Roi:
    """Formats a ROI so that it is inside the image.

    A ROI is a 4-list (or 4-tuple) where the 2 first entries are the (x, y) coordinates of the top-left corner,
    and the 2 last entries are the (x, y) coordinates of the bottom right corner.
    This function checks that the corners are well-ordered and inside the image.
    None is treated as min(=0) for the TL corner or max(=width or height) for the BR corner.

    Parameters
    ----------
    image : ndarray

    roi : Roi
        Initial ROI. Can be None, or [None, None, None, None]

    Returns
    -------
    roi : Roi
        The formatted ROI

    """
    if roi is None:
        roi = [None, None, None, None]  # TLx, TLy, BRx, BRy
    height, width = image.shape

    tlx, tly, brx, bry = roi
    if tlx is None:
        trace('format_roi: TLX not provided.')
        tlx = 0
    else:
        if not (0 <= tlx < width):
            warning(f'TLX="{tlx}" does not verify 0 <= TLX < width={width}. Its was overriden: TLX=0')
            tlx = 0

    if tly is None:
        trace('format_roi: TLX not provided.')
        tly = 0
    else:
        if not (0 <= tly < height):
            warning(f'TLY="{tly}" does not verify 0 <= TLY < height={height}. Its was overriden: TLY=0')
            tly = 0

    if brx is None:
        trace('format_roi: BRX not provided.')
        brx = None
    else:
        if not (tlx < brx <= width):
            warning(
                f'BRX="{brx}" does not verify TLX={tlx} < BRX <= width={width}. Its was overriden: BRX=None (=width)')
            brx = None

    if bry is None:
        trace('format_roi: BRY not provided.')
        bry = None
    else:
        if not (tly < bry <= height):
            warning(
                f'BRY="{bry}" does not verify TLY={tly} < BRY <= height={height}. Its was overriden: BRX=None (=height)')
            brx = None

    trace(f'format_roi: {roi} -> {[tlx, tly, brx, bry]}')
    return [tlx, tly, brx, bry]


def otsu_intraclass_variance(image:np.ndarray, threshold: Union[int, float]) -> float:
    """Otsu's intra-class variance.

    If all pixels are above or below the threshold, this will throw a warning that can safely be ignored.

    Parameters
    ----------
    image : ndarray
        The image

    threshold : float

    Returns
    -------
    variance : float

    """
    try:
        return np.nansum(
            [
                np.mean(cls) * np.var(image, where=cls)
                #   weight   ·  intra-class variance
                for cls in [image >= threshold, image < threshold]
            ]
        )
    except:
        return 0.
    # NaNs only arise if the class is empty, in which case the contribution should be zero, which `nansum` accomplishes.

#TODO: Remove this
def otsu_threshold_slower(image:np.ndarray) -> int:
    """Otsu's optimal threshold for an image.

    Computes `Otsu's intraclass variance <pypendentdrop.otsu_intraclass_variance>` for all integers 0-225 and returns the best threshold.

    Parameters
    ----------
    image : ndarray

    Returns
    -------
    thershold : int

    """
    test_tresholds = np.arange(255, dtype=float)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        otsu_variance = np.array([otsu_intraclass_variance(image, test_treshold) for test_treshold in test_tresholds])

    best_threshold_otsu = int(test_tresholds[np.argmin(otsu_variance)])

    return best_threshold_otsu

#TODO: Remove this
def otsu_threshold_faster(image:np.ndarray) -> int:
    # adapted from https://stackoverflow.com/questions/48213278/implementing-otsu-binarization-from-scratch-python
    # Because we did not want to import scikit.image of opencv
    pixel_number = image.shape[0] * image.shape[1]
    mean_weight = 1.0 / pixel_number
    his, bins = np.histogram(image, np.arange(0, 256 + 1))
    final_thresh = -1
    final_value = -1
    intensity_arr = np.arange(256)
    for t in bins[1:-1]:  # This goes from 1 to 254 uint8 range (Pretty sure wont be those values)
        pcb = np.sum(his[:t])
        pcf = np.sum(his[t:])
        Wb = pcb * mean_weight
        Wf = pcf * mean_weight

        mub = np.sum(intensity_arr[:t] * his[:t]) / float(pcb)
        muf = np.sum(intensity_arr[t:] * his[t:]) / float(pcf)
        #print mub, muf
        value = Wb * Wf * (mub - muf) ** 2

        if value > final_value:  # this is inefficient but only computed 254 times so should be ok
            final_thresh = t
            final_value = value

    return final_thresh


def threshold_otsu(image:np.ndarray) -> float:
    """Find the threshold value for a bimodal histogram using the Otsu method.

    If you have a distribution that is bimodal (AKA with two peaks, with a valley
    between them), then you can use this to find the location of that valley, that
    splits the distribution into two.

    From the SciKit Image threshold_otsu implementation:
    https://github.com/scikit-image/scikit-image/blob/70fa904eee9ef370c824427798302551df57afa1/skimage/filters/thresholding.py#L312
    """
    if image.min() == image.max():
        return float(image[0,0])
    counts, bin_edges = np.histogram(image.ravel(), bins=np.arange(image.min()-0.5, image.max()+.5+1, 1))
    bin_centers:np.ndarray = (bin_edges[1:] + bin_edges[:-1]) / 2
    # trace(f'bin centers: {bin_centers}') # a bit too much lol

    # class probabilities for all possible thresholds
    weight1 = np.cumsum(counts)
    weight2 = np.cumsum(counts[::-1])[::-1]
    # class means for all possible thresholds
    mean1 = np.cumsum(counts * bin_centers) / weight1
    mean2 = (np.cumsum((counts * bin_centers)[::-1]) / weight2[::-1])[::-1]

    # Clip ends to align class 1 and class 2 variables:
    # The last value of ``weight1``/``mean1`` should pair with zero values in
    # ``weight2``/``mean2``, which do not exist.
    variance12 = weight1[:-1] * weight2[1:] * (mean1[:-1] - mean2[1:]) ** 2

    threshold:float = float(bin_centers[np.argmax(variance12)])
    return threshold

def auto_threshold(image: np.ndarray, roi: Roi = None) -> float:
    """Finds the most appropriate threshold for the image.

    Trying to find Otsu's most appropriate threshold for the image, falling back to 127 it it fails.

    Parameters
    ----------
    image : ndarray
    roi : Roi, optional

    Returns
    -------
    threshold : float

    """
    roi = format_roi(image, roi=roi)
    trace(f"auto_threshold: Trying to find Otsu's optimal threshold")
    try:
        # threshold:int = otsu_threshold_slower(image[roi[0]:roi[2], roi[1]:roi[3]])
        # threshold: int = otsu_threshold_faster(image[roi[0]:roi[2], roi[1]:roi[3]])
        threshold: int = int(np.rint(threshold_otsu(image[roi[0]:roi[2], roi[1]:roi[3]])))
    except:
        threshold = 127
        error('Encountered an error while computing the best threshold.')
    trace(f'auto_threshold: Best threshold for the selected region of the image is {threshold}')
    return threshold


def detect_contourlines(image: np.ndarray, level: float, roi: Roi = None) -> List[np.ndarray]:
    """Returns all the closed lines enclosing regions in ``image`` that are above ``level``.

    Returns a collection of lines that each a contour of the level ``level`` of the image.
    Each line is in line-form, i.e. shape=(N,2).

    Parameters
    ----------
    image : ndarray
    level : float
    roi : Roi, optional

    Returns
    -------
    lines : array_like
        A collection of ndarrays of shape (N, 2).

    """
    trace('detect_contourlines: called')
    roi = format_roi(image, roi=roi)

    cont_gen = contour_generator(z=image[roi[1]:roi[3], roi[0]:roi[2]], line_type=LineType.Separate)  # quad_as_tri=True

    lines = cont_gen.lines(level)

    for i_line, line in enumerate(lines):
        lines[i_line] = np.array(line) + np.expand_dims(np.array(roi[:2]), 0)

    return lines


def detect_main_contour(image: np.ndarray, level: float, roi: Roi = None) -> np.ndarray:
    """Returns the main (longest) closed line enclosing a region in ``image`` that is above ``level``.

    Finds the longest of all `contour lines <pypendentdrop.detect_contourlines>` above a specific level,
    and returns its transposition, so that it is of shape (2, N).

    Parameters
    ----------
    image : ndarray
    level : float
    roi : Roi, optional

    Returns
    -------
    lines : ndarray
        An ndarray of shape (2, N).

    """
    lines = detect_contourlines(image, level, roi=roi)

    return np.array(lines[np.argmax([len(line) for line in lines])]).T



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ppd_mainwindow.ui'
##
## Created by: Qt User Interface Compiler version 6.8.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

# + : Replaced PySide6 or PyQt6 by pyqtgraph.Qt

from pyqtgraph.Qt.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from pyqtgraph.Qt.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from pyqtgraph.Qt.QtWidgets import (QAbstractSpinBox, QApplication, QCheckBox, QDoubleSpinBox,
    QFormLayout, QFrame, QGridLayout, QGroupBox,
    QHBoxLayout, QLabel, QLineEdit, QMainWindow,
    QMenuBar, QPushButton, QSizePolicy, QSpacerItem,
    QSpinBox, QStackedWidget, QTabWidget, QVBoxLayout,
    QWidget)

class Ui_PPD_MainWindow(object):
    def setupUi(self, PPD_MainWindow):
        if not PPD_MainWindow.objectName():
            PPD_MainWindow.setObjectName(u"PPD_MainWindow")
        PPD_MainWindow.resize(1619, 894)
        self.centralwidget = QWidget(PPD_MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.horizontalLayout = QHBoxLayout(self.centralwidget)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.displayStackedWidget = QStackedWidget(self.centralwidget)
        self.displayStackedWidget.setObjectName(u"displayStackedWidget")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.displayStackedWidget.sizePolicy().hasHeightForWidth())
        self.displayStackedWidget.setSizePolicy(sizePolicy)
        self.display_welcomePage = QWidget()
        self.display_welcomePage.setObjectName(u"display_welcomePage")
        sizePolicy.setHeightForWidth(self.display_welcomePage.sizePolicy().hasHeightForWidth())
        self.display_welcomePage.setSizePolicy(sizePolicy)
        self.verticalLayout_5 = QVBoxLayout(self.display_welcomePage)
        self.verticalLayout_5.setSpacing(6)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.verticalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.welcomeIcon = QLabel(self.display_welcomePage)
        self.welcomeIcon.setObjectName(u"welcomeIcon")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy1.setHorizontalStretch(1)
        sizePolicy1.setVerticalStretch(1)
        sizePolicy1.setHeightForWidth(self.welcomeIcon.sizePolicy().hasHeightForWidth())
        self.welcomeIcon.setSizePolicy(sizePolicy1)
        self.welcomeIcon.setScaledContents(True)
        self.welcomeIcon.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.verticalLayout_5.addWidget(self.welcomeIcon)

        self.displayStackedWidget.addWidget(self.display_welcomePage)

        self.horizontalLayout.addWidget(self.displayStackedWidget)

        self.analysisTabs = QTabWidget(self.centralwidget)
        self.analysisTabs.setObjectName(u"analysisTabs")
        self.tab_imagecontour = QWidget()
        self.tab_imagecontour.setObjectName(u"tab_imagecontour")
        self.verticalLayout_4 = QVBoxLayout(self.tab_imagecontour)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.imageSelectionGroupBox = QGroupBox(self.tab_imagecontour)
        self.imageSelectionGroupBox.setObjectName(u"imageSelectionGroupBox")
        self.formLayout = QFormLayout(self.imageSelectionGroupBox)
        self.formLayout.setObjectName(u"formLayout")
        self.formLayout.setFormAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignVCenter)
        self.imageFileLabel = QLabel(self.imageSelectionGroupBox)
        self.imageFileLabel.setObjectName(u"imageFileLabel")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.imageFileLabel)

        self.imageFileHLayout = QHBoxLayout()
        self.imageFileHLayout.setObjectName(u"imageFileHLayout")
        self.imageFileLineEdit = QLineEdit(self.imageSelectionGroupBox)
        self.imageFileLineEdit.setObjectName(u"imageFileLineEdit")

        self.imageFileHLayout.addWidget(self.imageFileLineEdit)

        self.imageFileBrowsePushButton = QPushButton(self.imageSelectionGroupBox)
        self.imageFileBrowsePushButton.setObjectName(u"imageFileBrowsePushButton")

        self.imageFileHLayout.addWidget(self.imageFileBrowsePushButton)


        self.formLayout.setLayout(0, QFormLayout.FieldRole, self.imageFileHLayout)

        self.subregionLabel = QLabel(self.imageSelectionGroupBox)
        self.subregionLabel.setObjectName(u"subregionLabel")

        self.formLayout.setWidget(1, QFormLayout.LabelRole, self.subregionLabel)

        self.subregionGridLayout = QGridLayout()
        self.subregionGridLayout.setObjectName(u"subregionGridLayout")
        self.descr1ROITL = QLabel(self.imageSelectionGroupBox)
        self.descr1ROITL.setObjectName(u"descr1ROITL")
        self.descr1ROITL.setAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignmentFlag.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)

        self.subregionGridLayout.addWidget(self.descr1ROITL, 0, 0, 1, 1)

        self.ROI_TL_x_spinBox = QSpinBox(self.imageSelectionGroupBox)
        self.ROI_TL_x_spinBox.setObjectName(u"ROI_TL_x_spinBox")
        self.ROI_TL_x_spinBox.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.ROI_TL_x_spinBox.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.ROI_TL_x_spinBox.setMaximum(9999)

        self.subregionGridLayout.addWidget(self.ROI_TL_x_spinBox, 0, 1, 1, 1)

        self.ROI_TL_y_spinBox = QSpinBox(self.imageSelectionGroupBox)
        self.ROI_TL_y_spinBox.setObjectName(u"ROI_TL_y_spinBox")
        self.ROI_TL_y_spinBox.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.ROI_TL_y_spinBox.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.ROI_TL_y_spinBox.setMaximum(99999)

        self.subregionGridLayout.addWidget(self.ROI_TL_y_spinBox, 0, 3, 1, 1)

        self.ROI_BR_y_spinBox = QSpinBox(self.imageSelectionGroupBox)
        self.ROI_BR_y_spinBox.setObjectName(u"ROI_BR_y_spinBox")
        self.ROI_BR_y_spinBox.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.ROI_BR_y_spinBox.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.ROI_BR_y_spinBox.setMaximum(99999)

        self.subregionGridLayout.addWidget(self.ROI_BR_y_spinBox, 1, 3, 1, 1)

        self.ROI_BR_x_spinBox = QSpinBox(self.imageSelectionGroupBox)
        self.ROI_BR_x_spinBox.setObjectName(u"ROI_BR_x_spinBox")
        self.ROI_BR_x_spinBox.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.ROI_BR_x_spinBox.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.ROI_BR_x_spinBox.setMaximum(9999)

        self.subregionGridLayout.addWidget(self.ROI_BR_x_spinBox, 1, 1, 1, 1)

        self.descr2ROITL = QLabel(self.imageSelectionGroupBox)
        self.descr2ROITL.setObjectName(u"descr2ROITL")

        self.subregionGridLayout.addWidget(self.descr2ROITL, 0, 2, 1, 1)

        self.descr3ROITL = QLabel(self.imageSelectionGroupBox)
        self.descr3ROITL.setObjectName(u"descr3ROITL")
        sizePolicy.setHeightForWidth(self.descr3ROITL.sizePolicy().hasHeightForWidth())
        self.descr3ROITL.setSizePolicy(sizePolicy)

        self.subregionGridLayout.addWidget(self.descr3ROITL, 0, 4, 1, 1)

        self.descr1ROIBR = QLabel(self.imageSelectionGroupBox)
        self.descr1ROIBR.setObjectName(u"descr1ROIBR")
        self.descr1ROIBR.setAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignmentFlag.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)

        self.subregionGridLayout.addWidget(self.descr1ROIBR, 1, 0, 1, 1)

        self.descr2ROIBR = QLabel(self.imageSelectionGroupBox)
        self.descr2ROIBR.setObjectName(u"descr2ROIBR")

        self.subregionGridLayout.addWidget(self.descr2ROIBR, 1, 2, 1, 1)

        self.descr3ROIBR = QLabel(self.imageSelectionGroupBox)
        self.descr3ROIBR.setObjectName(u"descr3ROIBR")

        self.subregionGridLayout.addWidget(self.descr3ROIBR, 1, 4, 1, 1)


        self.formLayout.setLayout(1, QFormLayout.FieldRole, self.subregionGridLayout)


        self.verticalLayout_4.addWidget(self.imageSelectionGroupBox)

        self.contourGroupBox = QGroupBox(self.tab_imagecontour)
        self.contourGroupBox.setObjectName(u"contourGroupBox")
        self.formLayout_2 = QFormLayout(self.contourGroupBox)
        self.formLayout_2.setObjectName(u"formLayout_2")
        self.autoThresholdCheckBox = QCheckBox(self.contourGroupBox)
        self.autoThresholdCheckBox.setObjectName(u"autoThresholdCheckBox")
        self.autoThresholdCheckBox.setChecked(True)

        self.formLayout_2.setWidget(0, QFormLayout.LabelRole, self.autoThresholdCheckBox)

        self.thresholdHLayout = QHBoxLayout()
        self.thresholdHLayout.setObjectName(u"thresholdHLayout")
        self.customThresholdLabel = QLabel(self.contourGroupBox)
        self.customThresholdLabel.setObjectName(u"customThresholdLabel")
        self.customThresholdLabel.setEnabled(False)

        self.thresholdHLayout.addWidget(self.customThresholdLabel)

        self.customThresholdSpinBox = QSpinBox(self.contourGroupBox)
        self.customThresholdSpinBox.setObjectName(u"customThresholdSpinBox")
        self.customThresholdSpinBox.setEnabled(False)
        self.customThresholdSpinBox.setMaximum(255)
        self.customThresholdSpinBox.setValue(127)

        self.thresholdHLayout.addWidget(self.customThresholdSpinBox)


        self.formLayout_2.setLayout(0, QFormLayout.FieldRole, self.thresholdHLayout)

        self.subpixelCheckBox = QCheckBox(self.contourGroupBox)
        self.subpixelCheckBox.setObjectName(u"subpixelCheckBox")
        self.subpixelCheckBox.setChecked(True)

        self.formLayout_2.setWidget(1, QFormLayout.LabelRole, self.subpixelCheckBox)

        self.smoothingCheckBox = QCheckBox(self.contourGroupBox)
        self.smoothingCheckBox.setObjectName(u"smoothingCheckBox")
        self.smoothingCheckBox.setChecked(True)

        self.formLayout_2.setWidget(2, QFormLayout.LabelRole, self.smoothingCheckBox)

        self.smoothingHLayout = QHBoxLayout()
        self.smoothingHLayout.setObjectName(u"smoothingHLayout")
        self.SmoothingDistanceLabel = QLabel(self.contourGroupBox)
        self.SmoothingDistanceLabel.setObjectName(u"SmoothingDistanceLabel")

        self.smoothingHLayout.addWidget(self.SmoothingDistanceLabel)

        self.smoothingDistanceSpinBox = QSpinBox(self.contourGroupBox)
        self.smoothingDistanceSpinBox.setObjectName(u"smoothingDistanceSpinBox")
        self.smoothingDistanceSpinBox.setMinimum(1)
        self.smoothingDistanceSpinBox.setMaximum(999)
        self.smoothingDistanceSpinBox.setSingleStep(2)
        self.smoothingDistanceSpinBox.setValue(11)

        self.smoothingHLayout.addWidget(self.smoothingDistanceSpinBox)


        self.formLayout_2.setLayout(2, QFormLayout.FieldRole, self.smoothingHLayout)


        self.verticalLayout_4.addWidget(self.contourGroupBox)

        self.analysisTabs.addTab(self.tab_imagecontour, "")
        self.tab_measure = QWidget()
        self.tab_measure.setObjectName(u"tab_measure")
        self.verticalLayout = QVBoxLayout(self.tab_measure)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.parametersGroupBox = QGroupBox(self.tab_measure)
        self.parametersGroupBox.setObjectName(u"parametersGroupBox")
        self.verticalLayout_3 = QVBoxLayout(self.parametersGroupBox)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.formLayout_3 = QFormLayout()
        self.formLayout_3.setObjectName(u"formLayout_3")
        self.pixelSizeLabel = QLabel(self.parametersGroupBox)
        self.pixelSizeLabel.setObjectName(u"pixelSizeLabel")

        self.formLayout_3.setWidget(1, QFormLayout.LabelRole, self.pixelSizeLabel)

        self.pixelSizeSpinbox = QDoubleSpinBox(self.parametersGroupBox)
        self.pixelSizeSpinbox.setObjectName(u"pixelSizeSpinbox")
        self.pixelSizeSpinbox.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.pixelSizeSpinbox.setDecimals(5)
        self.pixelSizeSpinbox.setMaximum(99999.000000000000000)
        self.pixelSizeSpinbox.setValue(0.000000000000000)

        self.formLayout_3.setWidget(1, QFormLayout.FieldRole, self.pixelSizeSpinbox)

        self.pixelDensityLabel = QLabel(self.parametersGroupBox)
        self.pixelDensityLabel.setObjectName(u"pixelDensityLabel")

        self.formLayout_3.setWidget(0, QFormLayout.LabelRole, self.pixelDensityLabel)

        self.pixelDensitySpinBox = QDoubleSpinBox(self.parametersGroupBox)
        self.pixelDensitySpinBox.setObjectName(u"pixelDensitySpinBox")
        self.pixelDensitySpinBox.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.pixelDensitySpinBox.setDecimals(3)
        self.pixelDensitySpinBox.setMaximum(999999.000000000000000)
        self.pixelDensitySpinBox.setValue(0.000000000000000)

        self.formLayout_3.setWidget(0, QFormLayout.FieldRole, self.pixelDensitySpinBox)


        self.verticalLayout_3.addLayout(self.formLayout_3)

        self.parametersLine = QFrame(self.parametersGroupBox)
        self.parametersLine.setObjectName(u"parametersLine")
        self.parametersLine.setFrameShape(QFrame.Shape.HLine)
        self.parametersLine.setFrameShadow(QFrame.Shadow.Sunken)

        self.verticalLayout_3.addWidget(self.parametersLine)

        self.parametersSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_3.addItem(self.parametersSpacer)

        self.autoGuessPushButton = QPushButton(self.parametersGroupBox)
        self.autoGuessPushButton.setObjectName(u"autoGuessPushButton")
        self.autoGuessPushButton.setEnabled(False)

        self.verticalLayout_3.addWidget(self.autoGuessPushButton)

        self.parametersGridLayout = QGridLayout()
        self.parametersGridLayout.setObjectName(u"parametersGridLayout")
        self.caplengthLabel = QLabel(self.parametersGroupBox)
        self.caplengthLabel.setObjectName(u"caplengthLabel")

        self.parametersGridLayout.addWidget(self.caplengthLabel, 4, 0, 1, 1)

        self.tipxCheckBox = QCheckBox(self.parametersGroupBox)
        self.tipxCheckBox.setObjectName(u"tipxCheckBox")
        self.tipxCheckBox.setChecked(True)

        self.parametersGridLayout.addWidget(self.tipxCheckBox, 1, 2, 1, 1)

        self.tipxLabel = QLabel(self.parametersGroupBox)
        self.tipxLabel.setObjectName(u"tipxLabel")

        self.parametersGridLayout.addWidget(self.tipxLabel, 1, 0, 1, 1)

        self.anglegSpinBox = QDoubleSpinBox(self.parametersGroupBox)
        self.anglegSpinBox.setObjectName(u"anglegSpinBox")
        self.anglegSpinBox.setDecimals(3)
        self.anglegSpinBox.setMinimum(-180.000000000000000)
        self.anglegSpinBox.setMaximum(180.000000000000000)
        self.anglegSpinBox.setSingleStep(0.500000000000000)
        self.anglegSpinBox.setValue(-180.000000000000000)

        self.parametersGridLayout.addWidget(self.anglegSpinBox, 0, 1, 1, 1)

        self.tipyLabel = QLabel(self.parametersGroupBox)
        self.tipyLabel.setObjectName(u"tipyLabel")

        self.parametersGridLayout.addWidget(self.tipyLabel, 2, 0, 1, 1)

        self.anglegCheckBox = QCheckBox(self.parametersGroupBox)
        self.anglegCheckBox.setObjectName(u"anglegCheckBox")
        self.anglegCheckBox.setChecked(True)

        self.parametersGridLayout.addWidget(self.anglegCheckBox, 0, 2, 1, 1)

        self.tipxSpinBox = QDoubleSpinBox(self.parametersGroupBox)
        self.tipxSpinBox.setObjectName(u"tipxSpinBox")
        self.tipxSpinBox.setDecimals(3)
        self.tipxSpinBox.setMaximum(99999.000000000000000)

        self.parametersGridLayout.addWidget(self.tipxSpinBox, 1, 1, 1, 1)

        self.tipySpinBox = QDoubleSpinBox(self.parametersGroupBox)
        self.tipySpinBox.setObjectName(u"tipySpinBox")
        self.tipySpinBox.setDecimals(3)
        self.tipySpinBox.setMaximum(99999.000000000000000)

        self.parametersGridLayout.addWidget(self.tipySpinBox, 2, 1, 1, 1)

        self.anglegLabel = QLabel(self.parametersGroupBox)
        self.anglegLabel.setObjectName(u"anglegLabel")

        self.parametersGridLayout.addWidget(self.anglegLabel, 0, 0, 1, 1)

        self.tipyCheckBox = QCheckBox(self.parametersGroupBox)
        self.tipyCheckBox.setObjectName(u"tipyCheckBox")
        self.tipyCheckBox.setChecked(True)

        self.parametersGridLayout.addWidget(self.tipyCheckBox, 2, 2, 1, 1)

        self.r0Label = QLabel(self.parametersGroupBox)
        self.r0Label.setObjectName(u"r0Label")

        self.parametersGridLayout.addWidget(self.r0Label, 3, 0, 1, 1)

        self.r0SpinBox = QDoubleSpinBox(self.parametersGroupBox)
        self.r0SpinBox.setObjectName(u"r0SpinBox")
        self.r0SpinBox.setDecimals(5)
        self.r0SpinBox.setMaximum(99999.000000000000000)
        self.r0SpinBox.setSingleStep(0.100000000000000)
        self.r0SpinBox.setValue(0.000000000000000)

        self.parametersGridLayout.addWidget(self.r0SpinBox, 3, 1, 1, 1)

        self.caplengthSpinBox = QDoubleSpinBox(self.parametersGroupBox)
        self.caplengthSpinBox.setObjectName(u"caplengthSpinBox")
        self.caplengthSpinBox.setDecimals(5)
        self.caplengthSpinBox.setMaximum(99999.000000000000000)
        self.caplengthSpinBox.setSingleStep(0.100000000000000)
        self.caplengthSpinBox.setValue(0.000000000000000)

        self.parametersGridLayout.addWidget(self.caplengthSpinBox, 4, 1, 1, 1)

        self.r0CheckBox = QCheckBox(self.parametersGroupBox)
        self.r0CheckBox.setObjectName(u"r0CheckBox")
        self.r0CheckBox.setChecked(True)

        self.parametersGridLayout.addWidget(self.r0CheckBox, 3, 2, 1, 1)

        self.caplengthCheckBox = QCheckBox(self.parametersGroupBox)
        self.caplengthCheckBox.setObjectName(u"caplengthCheckBox")
        self.caplengthCheckBox.setChecked(True)

        self.parametersGridLayout.addWidget(self.caplengthCheckBox, 4, 2, 1, 1)


        self.verticalLayout_3.addLayout(self.parametersGridLayout)

        self.optimizePushButton = QPushButton(self.parametersGroupBox)
        self.optimizePushButton.setObjectName(u"optimizePushButton")
        self.optimizePushButton.setEnabled(False)

        self.verticalLayout_3.addWidget(self.optimizePushButton)


        self.verticalLayout.addWidget(self.parametersGroupBox)

        self.physicsGroupBox = QGroupBox(self.tab_measure)
        self.physicsGroupBox.setObjectName(u"physicsGroupBox")
        self.verticalLayout_2 = QVBoxLayout(self.physicsGroupBox)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.formLayout_5 = QFormLayout()
        self.formLayout_5.setObjectName(u"formLayout_5")
        self.gLabel = QLabel(self.physicsGroupBox)
        self.gLabel.setObjectName(u"gLabel")

        self.formLayout_5.setWidget(0, QFormLayout.LabelRole, self.gLabel)

        self.gSpinBox = QDoubleSpinBox(self.physicsGroupBox)
        self.gSpinBox.setObjectName(u"gSpinBox")
        self.gSpinBox.setBaseSize(QSize(200, 30))
        self.gSpinBox.setReadOnly(False)
        self.gSpinBox.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.gSpinBox.setDecimals(3)
        self.gSpinBox.setMaximum(99999.000000000000000)
        self.gSpinBox.setValue(9.805999999999999)

        self.formLayout_5.setWidget(0, QFormLayout.FieldRole, self.gSpinBox)

        self.dSpinBox = QDoubleSpinBox(self.physicsGroupBox)
        self.dSpinBox.setObjectName(u"dSpinBox")
        self.dSpinBox.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.dSpinBox.setDecimals(3)
        self.dSpinBox.setMaximum(1000000000.000000000000000)
        self.dSpinBox.setSingleStep(0.100000000000000)
        self.dSpinBox.setValue(1.000000000000000)

        self.formLayout_5.setWidget(1, QFormLayout.FieldRole, self.dSpinBox)

        self.dLabel = QLabel(self.physicsGroupBox)
        self.dLabel.setObjectName(u"dLabel")

        self.formLayout_5.setWidget(1, QFormLayout.LabelRole, self.dLabel)


        self.verticalLayout_2.addLayout(self.formLayout_5)

        self.measureLine = QFrame(self.physicsGroupBox)
        self.measureLine.setObjectName(u"measureLine")
        self.measureLine.setFrameShape(QFrame.Shape.HLine)
        self.measureLine.setFrameShadow(QFrame.Shadow.Sunken)

        self.verticalLayout_2.addWidget(self.measureLine)

        self.measureSpacer = QSpacerItem(20, 139, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_2.addItem(self.measureSpacer)

        self.variablesFormLayout = QFormLayout()
        self.variablesFormLayout.setObjectName(u"variablesFormLayout")
        self.variablesFormLayout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.AllNonFixedFieldsGrow)
        self.variablesFormLayout.setLabelAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignVCenter)
        self.variablesFormLayout.setFormAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.gammaLabel = QLabel(self.physicsGroupBox)
        self.gammaLabel.setObjectName(u"gammaLabel")

        self.variablesFormLayout.setWidget(0, QFormLayout.LabelRole, self.gammaLabel)

        self.gammaSpinBox = QDoubleSpinBox(self.physicsGroupBox)
        self.gammaSpinBox.setObjectName(u"gammaSpinBox")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.gammaSpinBox.sizePolicy().hasHeightForWidth())
        self.gammaSpinBox.setSizePolicy(sizePolicy2)
        self.gammaSpinBox.setBaseSize(QSize(200, 30))
        self.gammaSpinBox.setReadOnly(True)
        self.gammaSpinBox.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.gammaSpinBox.setDecimals(3)
        self.gammaSpinBox.setMinimum(0.000000000000000)
        self.gammaSpinBox.setMaximum(99999.990000000005239)
        self.gammaSpinBox.setValue(0.000000000000000)

        self.variablesFormLayout.setWidget(0, QFormLayout.FieldRole, self.gammaSpinBox)

        self.bondLabel = QLabel(self.physicsGroupBox)
        self.bondLabel.setObjectName(u"bondLabel")

        self.variablesFormLayout.setWidget(1, QFormLayout.LabelRole, self.bondLabel)

        self.bondSpinBox = QDoubleSpinBox(self.physicsGroupBox)
        self.bondSpinBox.setObjectName(u"bondSpinBox")
        self.bondSpinBox.setBaseSize(QSize(200, 30))
        self.bondSpinBox.setReadOnly(True)
        self.bondSpinBox.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.bondSpinBox.setDecimals(3)
        self.bondSpinBox.setMaximum(99999.990000000005239)
        self.bondSpinBox.setValue(0.000000000000000)

        self.variablesFormLayout.setWidget(1, QFormLayout.FieldRole, self.bondSpinBox)


        self.verticalLayout_2.addLayout(self.variablesFormLayout)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_2.addItem(self.verticalSpacer)

        self.formLayout_4 = QFormLayout()
        self.formLayout_4.setObjectName(u"formLayout_4")
        self.worthingtonLabel = QLabel(self.physicsGroupBox)
        self.worthingtonLabel.setObjectName(u"worthingtonLabel")

        self.formLayout_4.setWidget(3, QFormLayout.LabelRole, self.worthingtonLabel)

        self.worthingtonSpinBox = QDoubleSpinBox(self.physicsGroupBox)
        self.worthingtonSpinBox.setObjectName(u"worthingtonSpinBox")
        self.worthingtonSpinBox.setReadOnly(True)
        self.worthingtonSpinBox.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.worthingtonSpinBox.setDecimals(3)
        self.worthingtonSpinBox.setMaximum(99999.990000000005239)
        self.worthingtonSpinBox.setValue(0.000000000000000)

        self.formLayout_4.setWidget(3, QFormLayout.FieldRole, self.worthingtonSpinBox)

        self.dropVolumeLabel = QLabel(self.physicsGroupBox)
        self.dropVolumeLabel.setObjectName(u"dropVolumeLabel")

        self.formLayout_4.setWidget(1, QFormLayout.LabelRole, self.dropVolumeLabel)

        self.dropVolumeSpinBox = QDoubleSpinBox(self.physicsGroupBox)
        self.dropVolumeSpinBox.setObjectName(u"dropVolumeSpinBox")
        self.dropVolumeSpinBox.setReadOnly(True)
        self.dropVolumeSpinBox.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.dropVolumeSpinBox.setDecimals(3)
        self.dropVolumeSpinBox.setMaximum(1000000.000000000000000)

        self.formLayout_4.setWidget(1, QFormLayout.FieldRole, self.dropVolumeSpinBox)

        self.needleDiameterSpinBox = QDoubleSpinBox(self.physicsGroupBox)
        self.needleDiameterSpinBox.setObjectName(u"needleDiameterSpinBox")
        self.needleDiameterSpinBox.setReadOnly(True)
        self.needleDiameterSpinBox.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.needleDiameterSpinBox.setDecimals(3)
        self.needleDiameterSpinBox.setMaximum(1000000.000000000000000)

        self.formLayout_4.setWidget(2, QFormLayout.FieldRole, self.needleDiameterSpinBox)

        self.needleDiameterLabel = QLabel(self.physicsGroupBox)
        self.needleDiameterLabel.setObjectName(u"needleDiameterLabel")

        self.formLayout_4.setWidget(2, QFormLayout.LabelRole, self.needleDiameterLabel)


        self.verticalLayout_2.addLayout(self.formLayout_4)


        self.verticalLayout.addWidget(self.physicsGroupBox)

        self.analysisTabs.addTab(self.tab_measure, "")

        self.horizontalLayout.addWidget(self.analysisTabs)

        PPD_MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(PPD_MainWindow)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 1619, 23))
        PPD_MainWindow.setMenuBar(self.menubar)

        self.retranslateUi(PPD_MainWindow)
        self.autoThresholdCheckBox.toggled.connect(self.customThresholdLabel.setDisabled)
        self.smoothingCheckBox.toggled.connect(self.smoothingDistanceSpinBox.setEnabled)
        self.smoothingCheckBox.toggled.connect(self.SmoothingDistanceLabel.setEnabled)
        self.autoThresholdCheckBox.toggled.connect(self.customThresholdSpinBox.setDisabled)

        self.analysisTabs.setCurrentIndex(1)


        QMetaObject.connectSlotsByName(PPD_MainWindow)
    # setupUi

    def retranslateUi(self, PPD_MainWindow):
        PPD_MainWindow.setWindowTitle(QCoreApplication.translate("PPD_MainWindow", u"MainWindow", None))
        self.welcomeIcon.setText(QCoreApplication.translate("PPD_MainWindow", u"Py Pendent Drop", None))
        self.imageSelectionGroupBox.setTitle(QCoreApplication.translate("PPD_MainWindow", u"Image selection", None))
        self.imageFileLabel.setText(QCoreApplication.translate("PPD_MainWindow", u"Image file", None))
        self.imageFileLineEdit.setInputMask("")
        self.imageFileLineEdit.setPlaceholderText(QCoreApplication.translate("PPD_MainWindow", u"Select an image file", None))
        self.imageFileBrowsePushButton.setText(QCoreApplication.translate("PPD_MainWindow", u"Browse", None))
        self.subregionLabel.setText(QCoreApplication.translate("PPD_MainWindow", u"Subregion of interest", None))
        self.descr1ROITL.setText(QCoreApplication.translate("PPD_MainWindow", u"TL corner: (", None))
        self.descr2ROITL.setText(QCoreApplication.translate("PPD_MainWindow", u", ", None))
        self.descr3ROITL.setText(QCoreApplication.translate("PPD_MainWindow", u")", None))
        self.descr1ROIBR.setText(QCoreApplication.translate("PPD_MainWindow", u"BR corner: (", None))
        self.descr2ROIBR.setText(QCoreApplication.translate("PPD_MainWindow", u", ", None))
        self.descr3ROIBR.setText(QCoreApplication.translate("PPD_MainWindow", u")", None))
        self.contourGroupBox.setTitle(QCoreApplication.translate("PPD_MainWindow", u"Drop contour detection", None))
        #if QT_CONFIG(tooltip)
        self.autoThresholdCheckBox.setToolTip(QCoreApplication.translate("PPD_MainWindow", u"Choose automatically the binarization threshold, using Otsu' s method.", None))
        #endif // QT_CONFIG(tooltip)
        self.autoThresholdCheckBox.setText(QCoreApplication.translate("PPD_MainWindow", u"Auto threshold", None))
        #if QT_CONFIG(tooltip)
        self.customThresholdLabel.setToolTip(QCoreApplication.translate("PPD_MainWindow", u"The binarization threshold.", None))
        #endif // QT_CONFIG(tooltip)
        self.customThresholdLabel.setText(QCoreApplication.translate("PPD_MainWindow", u"Threshold", None))
        #if QT_CONFIG(tooltip)
        self.subpixelCheckBox.setToolTip(QCoreApplication.translate("PPD_MainWindow", u"Subpixel edge extraction, adapted from R. Ngiam.", None))
        #endif // QT_CONFIG(tooltip)
        self.subpixelCheckBox.setText(QCoreApplication.translate("PPD_MainWindow", u"Use subpixel refinement", None))
        #if QT_CONFIG(tooltip)
        self.smoothingCheckBox.setToolTip(QCoreApplication.translate("PPD_MainWindow", u"Smooth the obtained profile to remove nonphysical high frequency spikes.", None))
        #endif // QT_CONFIG(tooltip)
        self.smoothingCheckBox.setText(QCoreApplication.translate("PPD_MainWindow", u"Smooth profile", None))
        #if QT_CONFIG(tooltip)
        self.SmoothingDistanceLabel.setToolTip(QCoreApplication.translate("PPD_MainWindow", u"The typical distance of which the contour is smoothed.", None))
        #endif // QT_CONFIG(tooltip)
        self.SmoothingDistanceLabel.setText(QCoreApplication.translate("PPD_MainWindow", u"Smoothing distance", None))
        self.smoothingDistanceSpinBox.setSuffix(QCoreApplication.translate("PPD_MainWindow", u" px", None))
        self.analysisTabs.setTabText(self.analysisTabs.indexOf(self.tab_imagecontour), QCoreApplication.translate("PPD_MainWindow", u"Drop shape characterization", None))
        self.parametersGroupBox.setTitle(QCoreApplication.translate("PPD_MainWindow", u"Parameters finding", None))
        self.pixelSizeLabel.setText(QCoreApplication.translate("PPD_MainWindow", u"Pixel spacing", None))
        self.pixelSizeSpinbox.setSpecialValueText(QCoreApplication.translate("PPD_MainWindow", u"Click to provide pixel spacing (in mm/px)", None))
        self.pixelSizeSpinbox.setSuffix(QCoreApplication.translate("PPD_MainWindow", u" mm/px", None))
        self.pixelDensityLabel.setText(QCoreApplication.translate("PPD_MainWindow", u"Pixel density", None))
        self.pixelDensitySpinBox.setSpecialValueText(QCoreApplication.translate("PPD_MainWindow", u"Click to provide pixel density (in px/mm) ", None))
        self.pixelDensitySpinBox.setSuffix(QCoreApplication.translate("PPD_MainWindow", u" px/mm", None))
        self.autoGuessPushButton.setText(QCoreApplication.translate("PPD_MainWindow", u"Auto-guess parameters", None))
        self.caplengthLabel.setText(QCoreApplication.translate("PPD_MainWindow", u"Capillary length", None))
        self.tipxCheckBox.setText(QCoreApplication.translate("PPD_MainWindow", u"Fit this parameter", None))
        self.tipxLabel.setText(QCoreApplication.translate("PPD_MainWindow", u"Tip x position", None))
        self.anglegSpinBox.setSpecialValueText(QCoreApplication.translate("PPD_MainWindow", u"-", None))
        self.anglegSpinBox.setSuffix(QCoreApplication.translate("PPD_MainWindow", u" deg", None))
        self.tipyLabel.setText(QCoreApplication.translate("PPD_MainWindow", u"Tip y position", None))
        self.anglegCheckBox.setText(QCoreApplication.translate("PPD_MainWindow", u"Fit this parameter", None))
        self.tipxSpinBox.setSpecialValueText(QCoreApplication.translate("PPD_MainWindow", u"-", None))
        self.tipxSpinBox.setSuffix(QCoreApplication.translate("PPD_MainWindow", u" px", None))
        self.tipySpinBox.setSpecialValueText(QCoreApplication.translate("PPD_MainWindow", u"-", None))
        self.tipySpinBox.setSuffix(QCoreApplication.translate("PPD_MainWindow", u" px", None))
        self.anglegLabel.setText(QCoreApplication.translate("PPD_MainWindow", u"Angle of gravity", None))
        self.tipyCheckBox.setText(QCoreApplication.translate("PPD_MainWindow", u"Fit this parameter", None))
        self.r0Label.setText(QCoreApplication.translate("PPD_MainWindow", u"Drop tip radius", None))
        self.r0SpinBox.setSpecialValueText(QCoreApplication.translate("PPD_MainWindow", u"-", None))
        self.r0SpinBox.setSuffix(QCoreApplication.translate("PPD_MainWindow", u" mm", None))
        self.caplengthSpinBox.setSpecialValueText(QCoreApplication.translate("PPD_MainWindow", u"-", None))
        self.caplengthSpinBox.setSuffix(QCoreApplication.translate("PPD_MainWindow", u" mm", None))
        self.r0CheckBox.setText(QCoreApplication.translate("PPD_MainWindow", u"Fit this parameter", None))
        self.caplengthCheckBox.setText(QCoreApplication.translate("PPD_MainWindow", u"Fit this parameter", None))
        self.optimizePushButton.setText(QCoreApplication.translate("PPD_MainWindow", u"Optimize (fit the checked parameters)", None))
        self.physicsGroupBox.setTitle(QCoreApplication.translate("PPD_MainWindow", u"Physical variables", None))
        self.gLabel.setText(QCoreApplication.translate("PPD_MainWindow", u"Acceleration of gravity", None))
        self.gSpinBox.setSuffix(QCoreApplication.translate("PPD_MainWindow", u" m/s\u00b2", None))
        self.dSpinBox.setSuffix(QCoreApplication.translate("PPD_MainWindow", u" kg/l", None))
        self.dLabel.setText(QCoreApplication.translate("PPD_MainWindow", u"Density difference \u0394\u03c1", None))
        self.gammaLabel.setText(QCoreApplication.translate("PPD_MainWindow", u"Surface tension gamma", None))
        self.gammaSpinBox.setSpecialValueText(QCoreApplication.translate("PPD_MainWindow", u"-", None))
        self.gammaSpinBox.setSuffix(QCoreApplication.translate("PPD_MainWindow", u" mN/m", None))
        self.bondLabel.setText(QCoreApplication.translate("PPD_MainWindow", u"Bond number", None))
        self.bondSpinBox.setSpecialValueText(QCoreApplication.translate("PPD_MainWindow", u"-", None))
        self.worthingtonLabel.setText(QCoreApplication.translate("PPD_MainWindow", u"Worthington number", None))
        self.worthingtonSpinBox.setSpecialValueText(QCoreApplication.translate("PPD_MainWindow", u"Not implemented yet", None))
        self.dropVolumeLabel.setText(QCoreApplication.translate("PPD_MainWindow", u"Drop volume", None))
        self.dropVolumeSpinBox.setSpecialValueText(QCoreApplication.translate("PPD_MainWindow", u"Not implemented yet", None))
        self.dropVolumeSpinBox.setSuffix(QCoreApplication.translate("PPD_MainWindow", u" mm\u00b3", None))
        self.needleDiameterSpinBox.setSpecialValueText(QCoreApplication.translate("PPD_MainWindow", u"Not implemented yet", None))
        self.needleDiameterSpinBox.setSuffix(QCoreApplication.translate("PPD_MainWindow", u" mm", None))
        self.needleDiameterLabel.setText(QCoreApplication.translate("PPD_MainWindow", u"Needle diameter", None))
        self.analysisTabs.setTabText(self.analysisTabs.indexOf(self.tab_measure), QCoreApplication.translate("PPD_MainWindow", u"Measurement", None))
    # retranslateUi




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import argparse
import pyqtgraph as pg
from .mainwindow import ppd_mainwindow

from .. import logfacility

def main():
    parser = argparse.ArgumentParser(
        prog='ppd_GUI',
        description='PyPendentDrop - GUI version',
        epilog=f'', add_help=True)
    parser.add_argument('-v', help='Verbosity (-v: info, -vv: logger.debug, -vvv: trace)', action="count", default=0)

    args = parser.parse_args()

    logfacility.set_verbose(args.v)

    app = pg.mkQApp("PyPendentDrop")

    mainwindow = ppd_mainwindow()

    mainwindow.show()

    sys.exit(pg.exec())
